# -*- Mode: Python -*-
"""
Low level term evaluation functions.
"""
cimport cython

cimport numpy as np
import numpy as np

from sfepy.discrete.common.extmods.mappings cimport (Mapping, CMapping)
from sfepy.discrete.common.extmods._fmfield cimport (FMField,
                                                     array2fmfield4,
                                                     array2fmfield3,
                                                     array2fmfield2,
                                                     array2fmfield1,
                                                     array2pint1,
                                                     array2pint2)

from sfepy.discrete.common.extmods.types cimport int32, float64, complex128

cdef extern from 'common.h':
    cdef void _errclear 'errclear'()

cdef extern from 'terms.h':
    cdef int32 _dq_state_in_qp \
         'dq_state_in_qp'(FMField *out, FMField *state, int32 offset,
                          FMField *bf,
                          int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _dq_grad \
         'dq_grad'(FMField *out, FMField *state, int32 offset,
                   Mapping *vg, int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _dq_div_vector \
         'dq_div_vector'(FMField *out, FMField *state, int32 offset,
                         Mapping *vg,
                         int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _d_volume_surface \
         'd_volume_surface'(FMField *out, FMField *in_,
                            Mapping *sg,
                            int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _di_surface_moment \
         'di_surface_moment'(FMField *out, FMField *in_,
                             Mapping *sg,
                             int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _dq_finite_strain_tl \
         'dq_finite_strain_tl'(FMField *mtxF, FMField *detF, FMField *vecCS,
                               FMField *trC, FMField *in2C, FMField *vecInvCS,
                               FMField *vecES,
                               FMField *state, int32 offset,
                               Mapping *vg,
                               int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _dq_finite_strain_ul \
         'dq_finite_strain_ul'(FMField *mtxF, FMField *detF, FMField *vecBS,
                               FMField *trB, FMField *in2B, FMField *vecES,
                               FMField *state, int32 offset,
                               Mapping *vg,
                               int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _dq_tl_finite_strain_surface \
         'dq_tl_finite_strain_surface'(FMField *mtxF, FMField *detF,
                                       FMField *mtxFI,
                                       FMField *state, int32 offset,
                                       Mapping *sg,
                                       int32 *fis, int32 nFa, int32 nFP,
                                       int32 *conn, int32 nEl, int32 nE)

    cdef int32 _dq_tl_he_stress_bulk \
         'dq_tl_he_stress_bulk'(FMField *out,FMField *mat,
                                FMField *detF, FMField *vecInvCS)

    cdef int32 _dq_ul_he_stress_bulk \
         'dq_ul_he_stress_bulk'(FMField *out,FMField *mat,
                                FMField *detF)

    cdef int32 _dq_tl_he_stress_bulk_active \
         'dq_tl_he_stress_bulk_active'(FMField *out,FMField *mat,
                                       FMField *detF, FMField *vecInvCS)

    cdef int32 _dq_tl_he_stress_neohook \
         'dq_tl_he_stress_neohook'(FMField *out, FMField *mat,
                                   FMField *detF, FMField *trC,
                                   FMField *vecInvCS)

    cdef int32 _dq_ul_he_stress_neohook \
         'dq_ul_he_stress_neohook'(FMField *out, FMField *mat,
                                   FMField *detF, FMField *trB, FMField *vecBS)

    cdef int32 _dq_tl_he_stress_mooney_rivlin \
         'dq_tl_he_stress_mooney_rivlin'(FMField *out, FMField *mat,
                                         FMField *detF, FMField *trC,
                                         FMField *vecInvCS, FMField *vecCS,
                                         FMField *in2C)

    cdef int32 _dq_ul_he_stress_mooney_rivlin \
         'dq_ul_he_stress_mooney_rivlin'(FMField *out, FMField *mat,
                                         FMField *detF, FMField *trB,
                                         FMField *vecBS, FMField *in2B)

    cdef int32 _dq_tl_he_tan_mod_bulk \
         'dq_tl_he_tan_mod_bulk'(FMField *out, FMField *mat,
                                 FMField *detF, FMField *vecInvCS)

    cdef int32 _dq_ul_he_tan_mod_bulk \
         'dq_ul_he_tan_mod_bulk'(FMField *out, FMField *mat, FMField *detF)

    cdef int32 _dq_tl_he_tan_mod_bulk_active \
         'dq_tl_he_tan_mod_bulk_active'(FMField *out, FMField *mat,
                                        FMField *detF, FMField *vecInvCS)

    cdef int32 _dq_tl_he_tan_mod_neohook \
         'dq_tl_he_tan_mod_neohook'(FMField *out, FMField *mat,
                                    FMField *detF, FMField *trC,
                                    FMField *vecInvCS)

    cdef int32 _dq_ul_he_tan_mod_neohook \
         'dq_ul_he_tan_mod_neohook'(FMField *out, FMField *mat,
                                    FMField *detF, FMField *trB,
                                    FMField *vecBS)

    cdef int32 _dq_tl_he_tan_mod_mooney_rivlin \
         'dq_tl_he_tan_mod_mooney_rivlin'(FMField *out, FMField *mat,
                                          FMField *detF, FMField *trC,
                                          FMField *vecInvCS, FMField *vecCS,
                                          FMField *in2C)

    cdef int32 _dq_ul_he_tan_mod_mooney_rivlin \
         'dq_ul_he_tan_mod_mooney_rivlin'(FMField *out, FMField *mat,
                                          FMField *detF, FMField *trB,
                                          FMField *vecBS, FMField *in2B)

    cdef int32 _dw_he_rtm \
         'dw_he_rtm'(FMField *out,
                     FMField *stress, FMField *tan_mod,
                     FMField *mtxF, FMField *detF,
                     Mapping *vg,
                     int32 isDiff, int32 mode_ul)

    cdef int32 _de_he_rtm \
         'de_he_rtm'(FMField *out,
                     FMField *stress, FMField *detF,
                     Mapping *vg,
                     int32 *elList, int32 elList_nRow,
                     int32 mode_ul)

    cdef int32 _dq_tl_stress_bulk_pressure \
         'dq_tl_stress_bulk_pressure'(FMField *out, FMField *pressure_qp,
                                      FMField *detF, FMField *vecInvCS)
    cdef int32 _dq_ul_stress_bulk_pressure \
         'dq_ul_stress_bulk_pressure'(FMField *out, FMField *pressure_qp,
                                      FMField *det)
    cdef int32 _dq_tl_tan_mod_bulk_pressure_u \
         'dq_tl_tan_mod_bulk_pressure_u'(FMField *out, FMField *pressure_qp,
                                         FMField *detF, FMField *vecInvCS)
    cdef int32 _dq_ul_tan_mod_bulk_pressure_u \
         'dq_ul_tan_mod_bulk_pressure_u'(FMField *out, FMField *pressure_qp,
                                         FMField *detF)

    cdef int32 _dw_tl_volume \
         'dw_tl_volume'(FMField *out, FMField *mtxF,
                        FMField *vecInvCS, FMField *detF,
                        Mapping *vgs, Mapping *vgv,
                        int32 transpose, int32 mode)
    cdef int32 _dw_ul_volume \
         'dw_ul_volume'(FMField *out, FMField *detF,
                        Mapping *vgs, Mapping *vgv,
                        int32 transpose, int32 mode)

    cdef int32 _dw_tl_diffusion \
         'dw_tl_diffusion'(FMField *out, FMField *pressure_grad,
                           FMField *mtxD, FMField *ref_porosity,
                           FMField *mtxF, FMField *detF,
                           Mapping *vg, int32 mode)

    cdef int32 _d_tl_surface_flux \
         "d_tl_surface_flux"( FMField *out, FMField *pressure_grad,
                              FMField *mtxD, FMField *ref_porosity,
                              FMField *mtxFI, FMField *detF,
                              Mapping *sg, int32 mode )

    cdef int32 _dw_tl_surface_traction \
         'dw_tl_surface_traction'(FMField *out, FMField *traction,
                                  FMField *detF, FMField *mtxFI,
                                  FMField *bf, Mapping *sg,
                                  int32 *fis, int32 nFa, int32 nFP,
                                  int32 mode)

    cdef int32 _d_tl_volume_surface \
         'd_tl_volume_surface'(FMField *out, FMField *coors,
                               FMField *detF, FMField *mtxFI,
                               FMField *bf, Mapping *sg,
                               int32 *conn, int32 nFa, int32 nFP)

    cdef int32 _dq_def_grad \
         'dq_def_grad'(FMField *out, FMField *state, Mapping *vg,
                       int32 *conn, int32 nEl, int32 nEP, int32 mode)

    cdef int32 _he_residuum_from_mtx \
         'he_residuum_from_mtx'(FMField *out, FMField *mtxD,
                                FMField *state,
                                int32 *conn, int32 nEl, int32 nEP,
                                int32 *elList, int32 elList_nRo)
    cdef int32 _he_eval_from_mtx \
         'he_eval_from_mtx'(FMField *out, FMField *mtxD,
                            FMField *stateV, FMField *stateU,
                            int32 *conn, int32 nEl, int32 nEP,
                            int32 *elList, int32 elList_nRo)

    cdef int32 _dw_laplace \
         'dw_laplace'(FMField *out, FMField *grad,
                      FMField *coef, Mapping *vg,
                      int32 isDiff)

    cdef int32 _d_laplace \
         'd_laplace'(FMField *out, FMField *gradP1, FMField *gradP2,
                     FMField *coef, Mapping *vg)
    cdef int32 _dw_diffusion \
         'dw_diffusion'(FMField *out, FMField *grad,
                        FMField *mtxD, Mapping *vg,
                        int32 isDiff)
    cdef int32 _d_diffusion \
         'd_diffusion'(FMField *out, FMField *gradP1, FMField *gradP2,
                       FMField *mtxD, Mapping *vg)
    cdef int32 _dw_diffusion_r \
         'dw_diffusion_r'(FMField *out, FMField *mtxD, Mapping *vg)
    cdef int32 _d_surface_flux \
         'd_surface_flux'(FMField *out, FMField *grad,
                          FMField *mtxD, Mapping *sg, int32 mode)
    cdef int32 _dw_surface_flux \
         'dw_surface_flux'(FMField *out, FMField *grad,
                           FMField *mat, FMField *bf, Mapping *sg,
                           int32 *fis, int32 nFa, int32 nFP, int32 mode)
    cdef int32 _dw_convect_v_grad_s \
         'dw_convect_v_grad_s'(FMField *out, FMField *val_v, FMField *grad_s,
                               Mapping *vvg, Mapping *svg,
                               int32 isDiff)

    cdef int32 _dw_lin_elastic \
         'dw_lin_elastic'(FMField *out, float64 coef, FMField *strain,
                          FMField *mtxD, Mapping *vg,
                          int32 isDiff)
    cdef int32 _d_lin_elastic \
         'd_lin_elastic'(FMField *out, float64 coef, FMField *strainV,
                         FMField *strainU, FMField *mtxD, Mapping *vg)

    cdef int32 _d_sd_lin_elastic \
         'd_sd_lin_elastic'(FMField *out, float64 coef, FMField *gradV,
                            FMField *gradU, FMField *gradW, FMField *mtxD,
                            Mapping *vg)

    cdef int32 _dw_lin_prestress \
         'dw_lin_prestress'(FMField *out, FMField *stress, Mapping *vg)

    cdef int32 _dw_lin_strain_fib \
         'dw_lin_strain_fib'(FMField *out, FMField *mtxD, FMField *mat,
                             Mapping *vg)

    cdef int32 _de_cauchy_strain \
         'de_cauchy_strain'(FMField *out, FMField *strain,
                            Mapping *vg, int32 mode)
    cdef int32 _de_cauchy_stress \
         'de_cauchy_stress'(FMField *out, FMField *strain,
                            FMField *mtxD,  Mapping *vg,
                            int32 mode)
    cdef int32 _dq_cauchy_strain \
         'dq_cauchy_strain'(FMField *out, FMField *state, int32 offset,
                            Mapping *vg,
                            int32 *conn, int32 nEl, int32 nEP)

    cdef int32 _dw_nonsym_elastic \
         'dw_nonsym_elastic'(FMField *out, FMField *grad, FMField *mtxD,
                             Mapping *vg, int32 isDiff)

    cdef int32 _dw_surface_ltr \
         'dw_surface_ltr'(FMField *out, FMField *traction, Mapping *sg)

    cdef int32 _dw_volume_lvf \
         'dw_volume_lvf'(FMField *out, FMField *forceQP, Mapping *vg)

    cdef int32 _dw_surface_v_dot_n_s \
         'dw_surface_v_dot_n_s'(FMField *out,
                                FMField *coef, FMField *val_qp,
                                Mapping *rsg,
                                Mapping *csg,
                                int32 isDiff)

    cdef int32 _dw_surface_s_v_dot_n \
         'dw_surface_s_v_dot_n'(FMField *out,
                                FMField *coef, FMField *val_qp,
                                Mapping *rsg,
                                Mapping *csg,
                                int32 isDiff)

    cdef int32 _dw_volume_dot_vector \
         'dw_volume_dot_vector'(FMField *out, FMField *coef, FMField *val_qp,
                                Mapping *rvg, Mapping *cvg,
                                int32 isDiff)

    cdef int32 _dw_volume_dot_scalar \
         'dw_volume_dot_scalar'(FMField *out, FMField *coef, FMField *val_qp,
                                Mapping *rvg, Mapping *cvg,
                                int32 isDiff)

    cdef int32 _dw_v_dot_grad_s_vw \
         'dw_v_dot_grad_s_vw'(FMField *out, FMField *coef, FMField *grad,
                              Mapping *vvg, Mapping *svg,
                              int32 isDiff)

    cdef int32 _dw_v_dot_grad_s_sw \
         'dw_v_dot_grad_s_sw'(FMField *out, FMField *coef, FMField *val_qp,
                              Mapping *vvg, Mapping *svg,
                              int32 isDiff)

    cdef int32 _term_ns_asm_div_grad \
         'term_ns_asm_div_grad'(FMField *out, FMField *grad,
                                FMField *viscosity, Mapping *vg,
                                int32 isDiff)

    cdef int32 _term_ns_asm_convect \
         'term_ns_asm_convect'(FMField *out, FMField *grad, FMField *state,
                               Mapping *vg, int32 isDiff)

    cdef int32 _dw_lin_convect \
         'dw_lin_convect'(FMField *out, FMField *grad, FMField *stateB,
                          Mapping *vg, int32 isDiff)

    cdef int32 _dw_div \
         'dw_div'(FMField *out, FMField *coef, FMField *div,
                  Mapping *svg, Mapping *vvg,
                  int32 isDiff)

    cdef int32 _dw_grad \
         'dw_grad'(FMField *out, FMField *coef, FMField *state,
                   Mapping *svg, Mapping *vvg,
                   int32 isDiff)

    cdef int32 _dw_st_pspg_c \
         'dw_st_pspg_c'(FMField *out,
                        FMField *stateB, FMField *stateU,
                        FMField *coef,
                        Mapping *vg_p, Mapping *vg_u,
                        int32 *conn, int32 nEl, int32 nEP,
                        int32 isDiff)

    cdef int32 _dw_st_supg_p \
         'dw_st_supg_p'(FMField *out,
                        FMField *stateB, FMField *gradP,
                        FMField *coef,
                        Mapping *vg_u, Mapping *vg_p,
                        int32 isDiff)

    cdef int32 _dw_st_supg_c \
         'dw_st_supg_c'(FMField *out,
                        FMField *stateB, FMField *stateU,
                        FMField *coef, Mapping *vg,
                        int32 *conn, int32 nEl, int32 nEP,
                        int32 isDiff)

    cdef int32 _dw_st_grad_div \
         'dw_st_grad_div'(FMField *out, FMField *div,
                          FMField *coef, Mapping *vg,
                          int32 isDiff)

    cdef int32 _dw_biot_grad \
         'dw_biot_grad'(FMField *out, float64 coef, FMField *pressure_qp,
                        FMField *mtxD, Mapping *svg, Mapping *vvg,
                        int32 isDiff)

    cdef int32 _dw_biot_div \
         'dw_biot_div'(FMField *out, float64 coef, FMField *strain,
                       FMField *mtxD, Mapping *svg, Mapping *vvg,
                       int32 isDiff)

    cdef int32 _d_biot_div \
         'd_biot_div'(FMField *out, float64 coef, FMField *state,
                      FMField *strain, FMField *mtxD, Mapping *vg)

    cdef int32 _dw_piezo_coupling \
         'dw_piezo_coupling'(FMField *out, FMField *strain,
                             FMField *charge_grad,
                             FMField *mtxG, Mapping *vg,
                             int32 mode)

    cdef int32 _d_piezo_coupling \
         'd_piezo_coupling'(FMField *out, FMField *strain,
                            FMField *charge_grad,
                            FMField *mtxG, Mapping *vg)

    cdef int32 _dw_electric_source \
         'dw_electric_source'(FMField *out, FMField *grad, FMField *coef,
                              Mapping *vg)

    cdef int32 _d_sd_diffusion \
         'd_sd_diffusion'(FMField *out,
                          FMField *grad_q, FMField *grad_p,
                          FMField *grad_w, FMField *div_w,
                          FMField *mtxD, Mapping *vg)

    cdef int32 _dw_adj_convect1 \
         'dw_adj_convect1'(FMField *out, FMField *stateW, FMField *gradU,
                           Mapping *vg, int32 isDiff)

    cdef int32 _dw_adj_convect2 \
         'dw_adj_convect2'(FMField *out, FMField *stateW, FMField *stateU,
                           Mapping *vg, int32 isDiff)

    cdef int32 _dw_st_adj_supg_c \
         'dw_st_adj_supg_c'(FMField *out, FMField *stateW,
                            FMField *stateU, FMField *gradU,
                            FMField *coef, Mapping *vg,
                            int32 *conn, int32 nEl, int32 nEP,
                            int32 isDiff)

    cdef int32 _dw_st_adj1_supg_p \
         'dw_st_adj1_supg_p'(FMField *out, FMField *stateW, FMField *gradP,
                             FMField *coef, Mapping *vg_w,
                             int32 *conn_w, int32 nEl_w, int32 nEP_w,
                             int32 isDiff)

    cdef int32 _dw_st_adj2_supg_p \
         'dw_st_adj2_supg_p'(FMField *out, FMField *gradU, FMField *stateR,
                             FMField *coef,
                             Mapping *vg_u, Mapping *vg_r,
                             int32 *conn_r, int32 nEl_r, int32 nEP_r,
                             int32 isDiff)

    cdef int32 _d_of_nsMinGrad \
         'd_of_nsMinGrad'(FMField *out, FMField *grad,
                          FMField *viscosity, Mapping *vg)

    cdef int32 _d_of_nsSurfMinDPress \
         'd_of_nsSurfMinDPress'(FMField *out, FMField *pressure,
                                float64 weight, float64 bpress,
                                Mapping *sg, int32 isDiff)

    cdef int32 _d_sd_div \
         'd_sd_div'(FMField *out, FMField *divU, FMField *gradU,
                    FMField *stateP, FMField *divMV, FMField *gradMV,
                    Mapping *vg_u, int32 mode)

    cdef int32 _d_sd_div_grad \
         'd_sd_div_grad'(FMField *out, FMField *gradU, FMField *gradW,
                         FMField *divMV, FMField *gradMV, FMField *viscosity,
                         Mapping *vg_u, int32 mode)

    cdef int32 _d_sd_convect \
         'd_sd_convect'(FMField *out, FMField *stateU, FMField *gradU,
                        FMField *stateW, FMField *divMV, FMField *gradMV,
                        Mapping *vg_u, int32 mode)

    cdef int32 _d_sd_volume_dot \
         'd_sd_volume_dot'(FMField *out, FMField *stateP, FMField *stateQ,
                           FMField *divMV, Mapping *vg, int32 mode)

    cdef int32 _d_sd_st_grad_div \
         'd_sd_st_grad_div'(FMField *out, FMField *divU, FMField *gradU,
                            FMField *divW, FMField *gradW, FMField *divMV,
                            FMField *gradMV, FMField *coef,
                            Mapping *vg_u, int32 mode)

    cdef int32 _d_sd_st_supg_c \
         'd_sd_st_supg_c'(FMField *out, FMField *stateB, FMField *gradU,
                          FMField *gradW, FMField *divMV, FMField *gradMV,
                          FMField *coef, Mapping *vg_u, int32 mode)

    cdef int32 _d_sd_st_pspg_c \
         'd_sd_st_pspg_c'(FMField *out, FMField *stateB, FMField *gradU,
                          FMField *gradR, FMField *divMV, FMField *gradMV,
                          FMField *coef, Mapping *vg_u, int32 mode)

    cdef int32 _d_sd_st_pspg_p \
         'd_sd_st_pspg_p'(FMField *out, FMField *gradR, FMField *gradP,
                          FMField *divMV, FMField *gradMV, FMField *coef,
                          Mapping *vg_p, int32 mode)

    cdef int32 _mulAB_integrate \
         'mulAB_integrate'(FMField *out,
                           FMField *A, FMField *B,
                           Mapping *vg, int32 mode)

    cdef int32 _actBfT \
         'actBfT'(FMField *out,
                  FMField *bf, FMField *A)

    cdef int32 _sym2nonsym \
         'sym2nonsym'(FMField *out,
                      FMField *A)

def errclear():
    _errclear()

def dq_state_in_qp(np.ndarray out not None,
                   np.ndarray state not None,
                   np.ndarray bf not None,
                   np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _state, _bf
    cdef (int32 *) _conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield1(_state, state)
    array2fmfield4(_bf, bf)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_state_in_qp(_out, _state, 0, _bf, _conn, n_el, n_ep)
    return ret

def dq_grad(np.ndarray out not None,
            np.ndarray state not None,
            CMapping cmap not None,
            np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _state
    cdef (int32 *) _conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_grad(_out, _state, 0, cmap.geo, _conn, n_el, n_ep)
    return ret

def dq_div_vector(np.ndarray out not None,
                  np.ndarray state not None,
                  CMapping cmap not None,
                  np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _state
    cdef (int32 *) _conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_div_vector(_out, _state, 0, cmap.geo, _conn, n_el, n_ep)
    return ret

def d_volume_surface(np.ndarray out not None,
                     np.ndarray in_ not None,
                     CMapping cmap not None,
                     np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _in_
    cdef (int32 *) _conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield2(_in_, in_)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _d_volume_surface(_out, _in_, cmap.geo, _conn, n_el, n_ep)
    return ret

def di_surface_moment(np.ndarray out not None,
                      np.ndarray in_ not None,
                      CMapping cmap not None,
                      np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _in_
    cdef (int32 *) _conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield2(_in_, in_)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _di_surface_moment(_out, _in_, cmap.geo, _conn, n_el, n_ep)
    return ret

def dq_finite_strain_tl(np.ndarray mtx_f not None,
                        np.ndarray det_f not None,
                        np.ndarray vec_cs not None,
                        np.ndarray tr_c not None,
                        np.ndarray in_2c not None,
                        np.ndarray vec_inv_cs not None,
                        np.ndarray vec_es not None,
                        np.ndarray state not None,
                        CMapping cmap not None,
                        np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _mtx_f, _det_f, _vec_cs, _tr_c, _in_2c
    cdef FMField[1] _vec_inv_cs, _vec_es, _state
    cdef (int32 *) _conn
    cdef int32 n_el, n_ep

    array2fmfield4(_mtx_f, mtx_f)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_cs, vec_cs)
    array2fmfield4(_tr_c, tr_c)
    array2fmfield4(_in_2c, in_2c)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)
    array2fmfield4(_vec_es, vec_es)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_finite_strain_tl(_mtx_f, _det_f, _vec_cs, _tr_c, _in_2c,
                               _vec_inv_cs, _vec_es, _state, 0, cmap.geo,
                               _conn, n_el, n_ep)
    if ret:
        raise ValueError('ccore error (see above)')

    return ret

def dq_finite_strain_ul(np.ndarray mtx_f not None,
                        np.ndarray det_f not None,
                        np.ndarray vec_bs not None,
                        np.ndarray tr_b not None,
                        np.ndarray in_2b not None,
                        np.ndarray vec_es not None,
                        np.ndarray state not None,
                        CMapping cmap not None,
                        np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _mtx_f, _det_f, _vec_bs, _tr_b, _in_2b
    cdef FMField[1] _vec_es, _state
    cdef int32 *_conn
    cdef int32 n_el, n_ep

    array2fmfield4(_mtx_f, mtx_f)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_bs, vec_bs)
    array2fmfield4(_tr_b, tr_b)
    array2fmfield4(_in_2b, in_2b)
    array2fmfield4(_vec_es, vec_es)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_finite_strain_ul(_mtx_f, _det_f, _vec_bs, _tr_b, _in_2b,
                               _vec_es, _state, 0, cmap.geo,
                               _conn, n_el, n_ep)
    if ret:
        raise ValueError('ccore error (see above)')

    return ret

def dq_tl_finite_strain_surface(np.ndarray mtx_f not None,
                                np.ndarray det_f not None,
                                np.ndarray mtx_fi not None,
                                np.ndarray state not None,
                                CMapping cmap not None,
                                np.ndarray fis not None,
                                np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _mtx_f, _det_f, _mtx_fi, _state
    cdef (int32 *) _conn, _fis
    cdef int32 n_el, n_ep, n_fa, n_fp

    array2fmfield4(_mtx_f, mtx_f)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_mtx_fi, mtx_fi)
    array2fmfield1(_state, state)
    array2pint2(&_fis, &n_fa, &n_fp, fis)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_tl_finite_strain_surface(_mtx_f, _det_f, _mtx_fi, _state, 0,
                                       cmap.geo,
                                       _fis, n_fa, n_fp, _conn, n_el, n_ep)
    if ret:
        raise ValueError('ccore error (see above)')

    return ret

def dq_tl_he_stress_bulk(np.ndarray out not None,
                         np.ndarray mat not None,
                         np.ndarray det_f not None,
                         np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_he_stress_bulk(_out, _mat, _det_f, _vec_inv_cs)
    return ret

def dq_ul_he_stress_bulk(np.ndarray out not None,
                         np.ndarray mat not None,
                         np.ndarray det_f not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)

    ret = _dq_ul_he_stress_bulk(_out, _mat, _det_f)
    return ret

def dq_tl_he_stress_bulk_active(np.ndarray out not None,
                                np.ndarray mat not None,
                                np.ndarray det_f not None,
                                np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_he_stress_bulk_active(_out, _mat, _det_f, _vec_inv_cs)
    return ret

def dq_tl_he_stress_neohook(np.ndarray out not None,
                            np.ndarray mat not None,
                            np.ndarray det_f not None,
                            np.ndarray tr_c not None,
                            np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_c, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_c, tr_c)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_he_stress_neohook(_out, _mat, _det_f, _tr_c, _vec_inv_cs)
    return ret

def dq_ul_he_stress_neohook(np.ndarray out not None,
                            np.ndarray mat not None,
                            np.ndarray det_f not None,
                            np.ndarray tr_b not None,
                            np.ndarray vec_bs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_b, _vec_bs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_b, tr_b)
    array2fmfield4(_vec_bs, vec_bs)

    ret = _dq_ul_he_stress_neohook(_out, _mat, _det_f, _tr_b, _vec_bs)
    return ret

def dq_tl_he_stress_mooney_rivlin(np.ndarray out not None,
                                  np.ndarray mat not None,
                                  np.ndarray det_f not None,
                                  np.ndarray tr_c not None,
                                  np.ndarray vec_inv_cs not None,
                                  np.ndarray vec_cs not None,
                                  np.ndarray in_2c not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_c, _vec_inv_cs
    cdef FMField[1] _vec_cs, _in_2c

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_c, tr_c)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)
    array2fmfield4(_vec_cs, vec_cs)
    array2fmfield4(_in_2c, in_2c)

    ret = _dq_tl_he_stress_mooney_rivlin(_out, _mat, _det_f, _tr_c,
                                         _vec_inv_cs, _vec_cs, _in_2c)
    return ret

def dq_ul_he_stress_mooney_rivlin(np.ndarray out not None,
                                  np.ndarray mat not None,
                                  np.ndarray det_f not None,
                                  np.ndarray tr_b not None,
                                  np.ndarray vec_bs not None,
                                  np.ndarray in_2b not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_b, _vec_bs, _in_2b

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_b, tr_b)
    array2fmfield4(_vec_bs, vec_bs)
    array2fmfield4(_in_2b, in_2b)

    ret = _dq_ul_he_stress_mooney_rivlin(_out, _mat, _det_f, _tr_b,
                                         _vec_bs, _in_2b)
    return ret

def dq_tl_he_tan_mod_bulk(np.ndarray out not None,
                          np.ndarray mat not None,
                          np.ndarray det_f not None,
                          np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_he_tan_mod_bulk(_out, _mat, _det_f, _vec_inv_cs)
    return ret

def dq_ul_he_tan_mod_bulk(np.ndarray out not None,
                          np.ndarray mat not None,
                          np.ndarray det_f not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)

    ret = _dq_ul_he_tan_mod_bulk(_out, _mat, _det_f)
    return ret

def dq_tl_he_tan_mod_bulk_active(np.ndarray out not None,
                                 np.ndarray mat not None,
                                 np.ndarray det_f not None,
                                 np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_he_tan_mod_bulk_active(_out, _mat, _det_f, _vec_inv_cs)
    return ret

def dq_tl_he_tan_mod_neohook(np.ndarray out not None,
                             np.ndarray mat not None,
                             np.ndarray det_f not None,
                             np.ndarray tr_c not None,
                             np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_c, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_c, tr_c)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_he_tan_mod_neohook(_out, _mat, _det_f, _tr_c, _vec_inv_cs)
    return ret

def dq_ul_he_tan_mod_neohook(np.ndarray out not None,
                             np.ndarray mat not None,
                             np.ndarray det_f not None,
                             np.ndarray tr_b not None,
                             np.ndarray vec_bs not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_b, _vec_bs

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_b, tr_b)
    array2fmfield4(_vec_bs, vec_bs)

    ret = _dq_ul_he_tan_mod_neohook(_out, _mat, _det_f, _tr_b, _vec_bs)
    return ret

def dq_tl_he_tan_mod_mooney_rivlin(np.ndarray out not None,
                                   np.ndarray mat not None,
                                   np.ndarray det_f not None,
                                   np.ndarray tr_c not None,
                                   np.ndarray vec_inv_cs not None,
                                   np.ndarray vec_cs not None,
                                   np.ndarray in_2c not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_c, _vec_inv_cs
    cdef FMField[1] _vec_cs, _in_2c

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_c, tr_c)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)
    array2fmfield4(_vec_cs, vec_cs)
    array2fmfield4(_in_2c, in_2c)

    ret = _dq_tl_he_tan_mod_mooney_rivlin(_out, _mat, _det_f, _tr_c,
                                          _vec_inv_cs, _vec_cs, _in_2c)
    return ret

def dq_ul_he_tan_mod_mooney_rivlin(np.ndarray out not None,
                                   np.ndarray mat not None,
                                   np.ndarray det_f not None,
                                   np.ndarray tr_b not None,
                                   np.ndarray vec_bs not None,
                                   np.ndarray in_2b not None):
    cdef int32 ret
    cdef FMField[1] _out, _mat, _det_f, _tr_b, _vec_bs, _in_2b

    array2fmfield4(_out, out)
    array2fmfield4(_mat, mat)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_tr_b, tr_b)
    array2fmfield4(_vec_bs, vec_bs)
    array2fmfield4(_in_2b, in_2b)

    ret = _dq_ul_he_tan_mod_mooney_rivlin(_out, _mat, _det_f, _tr_b,
                                          _vec_bs, _in_2b)
    return ret

def dw_he_rtm(np.ndarray out not None,
              np.ndarray stress not None,
              np.ndarray tan_mod not None,
              np.ndarray mtx_f not None,
              np.ndarray det_f not None,
              CMapping cmap not None,
              int32 is_diff, int32 mode_ul):
    cdef int32 ret
    cdef FMField[1] _out, _stress, _tan_mod, _mtx_f, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_stress, stress)
    array2fmfield4(_tan_mod, tan_mod)
    array2fmfield4(_mtx_f, mtx_f)
    array2fmfield4(_det_f, det_f)

    ret = _dw_he_rtm(_out, _stress, _tan_mod, _mtx_f, _det_f,
                     cmap.geo, is_diff, mode_ul)
    return ret

def de_he_rtm(np.ndarray out not None,
              np.ndarray stress not None,
              np.ndarray det_f not None,
              CMapping cmap not None,
              np.ndarray el_list not None,
              int32 mode_ul):
    cdef int32 ret
    cdef FMField[1] _out, _stress, _det_f
    cdef int32 *_el_list
    cdef int32 n_el

    array2fmfield4(_out, out)
    array2fmfield4(_stress, stress)
    array2fmfield4(_det_f, det_f)
    array2pint1(&_el_list, &n_el, el_list)

    ret = _de_he_rtm(_out, _stress, _det_f,
                     cmap.geo, _el_list, n_el, mode_ul)
    return ret

def dq_tl_stress_bulk_pressure(np.ndarray out not None,
                               np.ndarray pressure_qp not None,
                               np.ndarray det_f not None,
                               np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_qp, _det_f, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_qp, pressure_qp)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_stress_bulk_pressure(_out, _pressure_qp, _det_f, _vec_inv_cs)
    return ret

def dq_ul_stress_bulk_pressure(np.ndarray out not None,
                               np.ndarray pressure_qp not None,
                               np.ndarray det_f not None):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_qp, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_qp, pressure_qp)
    array2fmfield4(_det_f, det_f)

    ret = _dq_ul_stress_bulk_pressure(_out, _pressure_qp, _det_f)
    return ret

def dq_tl_tan_mod_bulk_pressure_u(np.ndarray out not None,
                                  np.ndarray pressure_qp not None,
                                  np.ndarray det_f not None,
                                  np.ndarray vec_inv_cs not None):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_qp, _det_f, _vec_inv_cs

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_qp, pressure_qp)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)

    ret = _dq_tl_tan_mod_bulk_pressure_u(_out, _pressure_qp, _det_f,
                                         _vec_inv_cs)
    return ret

def dq_ul_tan_mod_bulk_pressure_u(np.ndarray out not None,
                                  np.ndarray pressure_qp not None,
                                  np.ndarray det_f not None):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_qp, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_qp, pressure_qp)
    array2fmfield4(_det_f, det_f)

    ret = _dq_ul_tan_mod_bulk_pressure_u(_out, _pressure_qp, _det_f)
    return ret

def dw_tl_volume(np.ndarray out not None,
                 np.ndarray mtx_f not None,
                 np.ndarray vec_inv_cs not None,
                 np.ndarray det_f not None,
                 CMapping cmap_s not None,
                 CMapping cmap_v not None,
                 int32 transpose, int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _mtx_f, _vec_inv_cs, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_mtx_f, mtx_f)
    array2fmfield4(_vec_inv_cs, vec_inv_cs)
    array2fmfield4(_det_f, det_f)

    ret = _dw_tl_volume(_out, _mtx_f, _vec_inv_cs, _det_f,
                        cmap_s.geo, cmap_v.geo, transpose, mode)
    return ret

def dw_ul_volume(np.ndarray out not None,
                 np.ndarray det_f not None,
                 CMapping cmap_s not None,
                 CMapping cmap_v not None,
                 int32 transpose, int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_det_f, det_f)

    ret = _dw_ul_volume(_out, _det_f, cmap_s.geo, cmap_v.geo, transpose, mode)
    return ret

def dw_tl_diffusion(np.ndarray out not None,
                    np.ndarray pressure_grad not None,
                    np.ndarray mtx_d not None,
                    np.ndarray ref_porosity not None,
                    np.ndarray mtx_f not None,
                    np.ndarray det_f not None,
                    CMapping cmap not None,
                    int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_grad, _mtx_d, _ref_porosity
    cdef FMField[1] _mtx_f, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_grad, pressure_grad)
    array2fmfield4(_mtx_d, mtx_d)
    array2fmfield4(_ref_porosity, ref_porosity)
    array2fmfield4(_mtx_f, mtx_f)
    array2fmfield4(_det_f, det_f)

    ret = _dw_tl_diffusion(_out, _pressure_grad, _mtx_d, _ref_porosity,
                           _mtx_f, _det_f, cmap.geo, mode)
    return ret

def d_tl_surface_flux(np.ndarray out not None,
                      np.ndarray pressure_grad not None,
                      np.ndarray mtx_d not None,
                      np.ndarray ref_porosity not None,
                      np.ndarray mtx_fi not None,
                      np.ndarray det_f not None,
                      CMapping cmap not None,
                      int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_grad, _mtx_d, _ref_porosity
    cdef FMField[1] _mtx_fi, _det_f

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_grad, pressure_grad)
    array2fmfield4(_mtx_d, mtx_d)
    array2fmfield4(_ref_porosity, ref_porosity)
    array2fmfield4(_mtx_fi, mtx_fi)
    array2fmfield4(_det_f, det_f)

    ret = _d_tl_surface_flux(_out, _pressure_grad, _mtx_d, _ref_porosity,
                             _mtx_fi, _det_f, cmap.geo, mode)
    return ret

def dw_tl_surface_traction(np.ndarray out not None,
                           np.ndarray traction not None,
                           np.ndarray det_f not None,
                           np.ndarray mtx_fi not None,
                           np.ndarray bf not None,
                           CMapping cmap not None,
                           np.ndarray fis not None,
                           int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _traction, _det_f, _mtx_fi, _bf
    cdef int32 *_fis
    cdef int32 n_fa, n_fp

    array2fmfield4(_out, out)
    array2fmfield4(_traction, traction)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_mtx_fi, mtx_fi)
    array2fmfield4(_bf, bf)
    array2pint2(&_fis, &n_fa, &n_fp, fis)

    ret = _dw_tl_surface_traction(_out, _traction, _det_f, _mtx_fi, _bf,
                                       cmap.geo, _fis, n_fa, n_fp, mode)
    return ret

def d_tl_volume_surface(np.ndarray out not None,
                        np.ndarray coors not None,
                        np.ndarray det_f not None,
                        np.ndarray mtx_fi not None,
                        np.ndarray bf not None,
                        CMapping cmap not None,
                        np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _coors, _det_f, _mtx_fi, _bf
    cdef int32 *_conn
    cdef int32 n_fa, n_fp

    array2fmfield4(_out, out)
    array2fmfield2(_coors, coors)
    array2fmfield4(_det_f, det_f)
    array2fmfield4(_mtx_fi, mtx_fi)
    array2fmfield4(_bf, bf)
    array2pint2(&_conn, &n_fa, &n_fp, conn)

    ret = _d_tl_volume_surface(_out, _coors, _det_f, _mtx_fi, _bf,
                               cmap.geo, _conn, n_fa, n_fp)
    return ret

def dq_def_grad(np.ndarray out not None,
                np.ndarray state not None,
                CMapping cmap not None,
                np.ndarray conn not None,
                int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _state
    cdef int32 *_conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_def_grad(_out, _state, cmap.geo, _conn, n_el, n_ep, mode)
    return ret

def he_residuum_from_mtx(np.ndarray out not None,
                         np.ndarray mtx_d not None,
                         np.ndarray state not None,
                         np.ndarray conn not None,
                         np.ndarray el_list not None):
    cdef int32 ret
    cdef FMField[1] _out, _mtx_d, _state
    cdef (int32 *) _conn, _el_list
    cdef int32 n_el, n_ep, n_el2

    array2fmfield4(_out, out)
    array2fmfield4(_mtx_d, mtx_d)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)
    array2pint1(&_el_list, &n_el2, el_list)

    ret = _he_residuum_from_mtx(_out, _mtx_d, _state,
                                _conn, n_el, n_ep, _el_list, n_el2)
    return ret

def he_eval_from_mtx(np.ndarray out not None,
                     np.ndarray mtx_d not None,
                     np.ndarray state_v not None,
                     np.ndarray state_u not None,
                     np.ndarray conn not None,
                     np.ndarray el_list not None):
    cdef int32 ret
    cdef FMField[1] _out, _mtx_d, _state_v, _state_u
    cdef (int32 *) _conn, _el_list
    cdef int32 n_el, n_ep, n_el2

    array2fmfield4(_out, out)
    array2fmfield4(_mtx_d, mtx_d)
    array2fmfield1(_state_v, state_v)
    array2fmfield1(_state_u, state_u)
    array2pint2(&_conn, &n_el, &n_ep, conn)
    array2pint1(&_el_list, &n_el2, el_list)

    ret = _he_eval_from_mtx(_out, _mtx_d, _state_v, _state_u,
                            _conn, n_el, n_ep, _el_list, n_el2)
    return ret

def dw_laplace(np.ndarray out not None,
               np.ndarray grad not None,
               np.ndarray coef not None,
               CMapping cmap not None,
               int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_coef, coef)

    ret = _dw_laplace(_out, _grad, _coef, cmap.geo, is_diff)
    return ret

def d_laplace(np.ndarray out not None,
              np.ndarray grad_p1 not None,
              np.ndarray grad_p2 not None,
              np.ndarray coef not None,
              CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _grad_p1, _grad_p2, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_grad_p1, grad_p1)
    array2fmfield4(_grad_p2, grad_p2)
    array2fmfield4(_coef, coef)

    ret = _d_laplace(_out, _grad_p1, _grad_p2, _coef, cmap.geo)
    return ret

def dw_diffusion(np.ndarray out not None,
                 np.ndarray grad not None,
                 np.ndarray mtx_d not None,
                 CMapping cmap not None,
                 int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _dw_diffusion(_out, _grad, _mtx_d, cmap.geo, is_diff)
    return ret

def d_diffusion(np.ndarray out not None,
                np.ndarray grad_p1 not None,
                np.ndarray grad_p2 not None,
                np.ndarray mtx_d not None,
                CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _grad_p1, _grad_p2, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_grad_p1, grad_p1)
    array2fmfield4(_grad_p2, grad_p2)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _d_diffusion(_out, _grad_p1, _grad_p2, _mtx_d, cmap.geo)
    return ret

def dw_diffusion_r(np.ndarray out not None,
                   np.ndarray mtx_d not None,
                   CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _dw_diffusion_r(_out, _mtx_d, cmap.geo)
    return ret

def d_surface_flux(np.ndarray out not None,
                   np.ndarray grad not None,
                   np.ndarray mtx_d not None,
                   CMapping cmap not None,
                   int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _d_surface_flux(_out, _grad, _mtx_d, cmap.geo, mode)
    return ret

def dw_surface_flux(np.ndarray out not None,
                    np.ndarray grad not None,
                    np.ndarray mat not None,
                    np.ndarray bf not None,
                    CMapping cmap not None,
                    np.ndarray fis not None,
                    int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _mat, _bf
    cdef int32 *_fis
    cdef int32 n_fa, n_fp

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_mat, mat)
    array2fmfield4(_bf, bf)
    array2pint2(&_fis, &n_fa, &n_fp, fis)

    ret = _dw_surface_flux(_out, _grad, _mat, _bf,
                           cmap.geo, _fis, n_fa, n_fp, mode)
    return ret

def dw_convect_v_grad_s(np.ndarray out not None,
                        np.ndarray val_v not None,
                        np.ndarray grad_s not None,
                        CMapping cmap_v not None,
                        CMapping cmap_s not None,
                        int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _val_v, _grad_s

    array2fmfield4(_out, out)
    array2fmfield4(_val_v, val_v)
    array2fmfield4(_grad_s, grad_s)

    ret = _dw_convect_v_grad_s(_out, _val_v, _grad_s,
                               cmap_v.geo, cmap_s.geo, is_diff)
    return ret

def dw_lin_elastic(np.ndarray out not None,
                   float64 coef,
                   np.ndarray strain not None,
                   np.ndarray mtx_d not None,
                   CMapping cmap not None,
                   int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _strain, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_strain, strain)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _dw_lin_elastic(_out, coef, _strain, _mtx_d, cmap.geo, is_diff)
    return ret

def d_lin_elastic(np.ndarray out not None,
                  float64 coef,
                  np.ndarray strain_v not None,
                  np.ndarray strain_u not None,
                  np.ndarray mtx_d not None,
                  CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _strain_u, _strain_v, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_strain_u, strain_u)
    array2fmfield4(_strain_v, strain_v)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _d_lin_elastic(_out, coef, _strain_u, _strain_v, _mtx_d, cmap.geo)
    return ret

def d_sd_lin_elastic(np.ndarray out not None,
                     float64 coef,
                     np.ndarray grad_v not None,
                     np.ndarray grad_u not None,
                     np.ndarray grad_w not None,
                     np.ndarray mtx_d not None,
                     CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _grad_u, _grad_v, _grad_w, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_grad_v, grad_v)
    array2fmfield4(_grad_w, grad_w)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _d_sd_lin_elastic(_out, coef, _grad_u, _grad_v, _grad_w,
                            _mtx_d, cmap.geo)
    return ret

def dw_lin_prestress(np.ndarray out not None,
                     np.ndarray stress not None,
                     CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _stress

    array2fmfield4(_out, out)
    array2fmfield4(_stress, stress)

    ret = _dw_lin_prestress(_out, _stress, cmap.geo)
    return ret

def dw_lin_strain_fib(np.ndarray out not None,
                      np.ndarray mtx_d not None,
                      np.ndarray mat not None,
                      CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _mtx_d, _mat

    array2fmfield4(_out, out)
    array2fmfield4(_mtx_d, mtx_d)
    array2fmfield4(_mat, mat)

    ret = _dw_lin_strain_fib(_out, _mtx_d, _mat, cmap.geo)
    return ret

def de_cauchy_strain(np.ndarray out not None,
                     np.ndarray strain not None,
                     CMapping cmap not None,
                     int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _strain

    array2fmfield4(_out, out)
    array2fmfield4(_strain, strain)

    ret = _de_cauchy_strain(_out, _strain, cmap.geo, mode)
    return ret

def de_cauchy_stress(np.ndarray out not None,
                     np.ndarray strain not None,
                     np.ndarray mtx_d not None,
                     CMapping cmap not None,
                     int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _strain, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_strain, strain)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _de_cauchy_stress(_out, _strain, _mtx_d, cmap.geo, mode)
    return ret

def dq_cauchy_strain(np.ndarray out not None,
                     np.ndarray state not None,
                     CMapping cmap not None,
                     np.ndarray conn not None):
    cdef int32 ret
    cdef FMField[1] _out, _state
    cdef int32 *_conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield1(_state, state)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dq_cauchy_strain(_out, _state, 0, cmap.geo, _conn, n_el, n_ep)
    return ret

def dw_nonsym_elastic(np.ndarray out not None,
                      np.ndarray grad not None,
                      np.ndarray mtx_d not None,
                      CMapping cmap not None,
                      int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _dw_nonsym_elastic(_out, _grad, _mtx_d, cmap.geo, is_diff)
    return ret

def dw_surface_ltr(np.ndarray out not None,
                   np.ndarray traction not None,
                   CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _traction

    array2fmfield4(_out, out)
    array2fmfield4(_traction, traction)

    ret = _dw_surface_ltr(_out, _traction, cmap.geo)
    return ret

def dw_volume_lvf(np.ndarray out not None,
                  np.ndarray force_qp not None,
                  CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _force_qp

    array2fmfield4(_out, out)
    array2fmfield4(_force_qp, force_qp)

    ret = _dw_volume_lvf(_out, _force_qp, cmap.geo)
    return ret

def dw_surface_v_dot_n_s(np.ndarray out not None,
                         np.ndarray coef not None,
                         np.ndarray val_qp not None,
                         CMapping rcmap not None,
                         CMapping ccmap not None,
                         int32 is_diff):

    cdef int32 ret
    cdef FMField[1] _out, _coef, _val_qp

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_val_qp, val_qp)

    ret = _dw_surface_v_dot_n_s(_out, _coef, _val_qp,
                                rcmap.geo, ccmap.geo, is_diff)
    return ret

def dw_surface_s_v_dot_n(np.ndarray out not None,
                         np.ndarray coef not None,
                         np.ndarray val_qp not None,
                         CMapping rcmap not None,
                         CMapping ccmap not None,
                         int32 is_diff):

    cdef int32 ret
    cdef FMField[1] _out, _coef, _val_qp

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_val_qp, val_qp)

    ret = _dw_surface_s_v_dot_n(_out, _coef, _val_qp,
                                rcmap.geo, ccmap.geo, is_diff)
    return ret

def dw_volume_dot_vector(np.ndarray out not None,
                         np.ndarray coef not None,
                         np.ndarray val_qp not None,
                         CMapping rcmap not None,
                         CMapping ccmap not None,
                         int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _coef, _val_qp

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_val_qp, val_qp)

    ret = _dw_volume_dot_vector(_out, _coef, _val_qp,
                                rcmap.geo, ccmap.geo, is_diff)
    return ret

def dw_volume_dot_scalar(np.ndarray out not None,
                         np.ndarray coef not None,
                         np.ndarray val_qp not None,
                         CMapping rcmap not None,
                         CMapping ccmap not None,
                         int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _coef, _val_qp

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_val_qp, val_qp)

    ret = _dw_volume_dot_scalar(_out, _coef, _val_qp,
                                rcmap.geo, ccmap.geo, is_diff)
    return ret

def dw_v_dot_grad_s_vw(np.ndarray out not None,
                       np.ndarray coef not None,
                       np.ndarray grad not None,
                       CMapping cmap_v not None,
                       CMapping cmap_s not None,
                       int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _coef, _grad

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_grad, grad)

    ret = _dw_v_dot_grad_s_vw(_out, _coef, _grad,
                              cmap_v.geo, cmap_s.geo, is_diff)
    return ret

def dw_v_dot_grad_s_sw(np.ndarray out not None,
                       np.ndarray coef not None,
                       np.ndarray val_qp not None,
                       CMapping cmap_v not None,
                       CMapping cmap_s not None,
                       int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _coef, _val_qp

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_val_qp, val_qp)

    ret = _dw_v_dot_grad_s_sw(_out, _coef, _val_qp,
                              cmap_v.geo, cmap_s.geo, is_diff)
    return ret

def term_ns_asm_div_grad(np.ndarray out not None,
                         np.ndarray grad not None,
                         np.ndarray viscosity not None,
                         CMapping cmap not None,
                         int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _viscosity

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_viscosity, viscosity)

    ret = _term_ns_asm_div_grad(_out, _grad, _viscosity, cmap.geo, is_diff)
    return ret

def term_ns_asm_convect(np.ndarray out not None,
                        np.ndarray grad not None,
                        np.ndarray state not None,
                        CMapping cmap not None,
                        int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _state

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_state, state)

    ret = _term_ns_asm_convect(_out, _grad, _state, cmap.geo, is_diff)
    return ret

def dw_lin_convect(np.ndarray out not None,
                   np.ndarray grad not None,
                   np.ndarray state_b not None,
                   CMapping cmap not None,
                   int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _state_b

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_state_b, state_b)

    ret = _dw_lin_convect(_out, _grad, _state_b, cmap.geo, is_diff)
    return ret

def dw_div(np.ndarray out not None,
           np.ndarray coef not None,
           np.ndarray div not None,
           CMapping cmap_s not None,
           CMapping cmap_v not None,
           int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _coef, _div

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_div, div)

    ret = _dw_div(_out, _coef, _div, cmap_s.geo, cmap_v.geo, is_diff)
    return ret

def dw_grad(np.ndarray out not None,
            np.ndarray coef not None,
            np.ndarray state not None,
            CMapping cmap_s not None,
            CMapping cmap_v not None,
            int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _coef, _state

    array2fmfield4(_out, out)
    array2fmfield4(_coef, coef)
    array2fmfield4(_state, state)

    ret = _dw_grad(_out, _coef, _state, cmap_s.geo, cmap_v.geo, is_diff)
    return ret

def dw_st_pspg_c(np.ndarray out not None,
                 np.ndarray state_b not None,
                 np.ndarray state_u not None,
                 np.ndarray coef not None,
                 CMapping cmap_p not None,
                 CMapping cmap_u not None,
                 np.ndarray conn not None,
                 int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_b, _state_u, _coef
    cdef int32 *_conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield4(_state_b, state_b)
    array2fmfield1(_state_u, state_u)
    array2fmfield4(_coef, coef)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dw_st_pspg_c(_out, _state_b, _state_u, _coef,
                        cmap_p.geo, cmap_u.geo, _conn, n_el, n_ep, is_diff)
    return ret

def dw_st_supg_p(np.ndarray out not None,
                 np.ndarray state_b not None,
                 np.ndarray grad_p not None,
                 np.ndarray coef not None,
                 CMapping cmap_u not None,
                 CMapping cmap_p not None,
                 int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_b, _grad_p, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_state_b, state_b)
    array2fmfield4(_grad_p, grad_p)
    array2fmfield4(_coef, coef)

    ret = _dw_st_supg_p(_out, _state_b, _grad_p, _coef,
                        cmap_u.geo, cmap_p.geo, is_diff)
    return ret

def dw_st_supg_c(np.ndarray out not None,
                 np.ndarray state_b not None,
                 np.ndarray state_u not None,
                 np.ndarray coef not None,
                 CMapping cmap not None,
                 np.ndarray conn not None,
                 int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_b, _state_u, _coef
    cdef int32 *_conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield4(_state_b, state_b)
    array2fmfield1(_state_u, state_u)
    array2fmfield4(_coef, coef)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dw_st_supg_c(_out, _state_b, _state_u, _coef,
                        cmap.geo, _conn, n_el, n_ep, is_diff)
    return ret

def dw_st_grad_div(np.ndarray out not None,
                   np.ndarray div not None,
                   np.ndarray coef not None,
                   CMapping cmap not None,
                   int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _div, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_div, div)
    array2fmfield4(_coef, coef)

    ret = _dw_st_grad_div(_out, _div, _coef, cmap.geo, is_diff)
    return ret

def dw_biot_grad(np.ndarray out not None,
                 float64 coef,
                 np.ndarray pressure_qp not None,
                 np.ndarray mtx_d not None,
                 CMapping cmap_s not None,
                 CMapping cmap_v not None,
                 int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _pressure_qp, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_pressure_qp, pressure_qp)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _dw_biot_grad(_out, coef, _pressure_qp, _mtx_d,
                        cmap_s.geo, cmap_v.geo, is_diff)
    return ret

def dw_biot_div(np.ndarray out not None,
                float64 coef,
                np.ndarray strain not None,
                np.ndarray mtx_d not None,
                CMapping cmap_s not None,
                CMapping cmap_v not None,
                int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _strain, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_strain, strain)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _dw_biot_div(_out, coef, _strain, _mtx_d,
                       cmap_s.geo, cmap_v.geo, is_diff)
    return ret

def d_biot_div(np.ndarray out not None,
               float64 coef,
               np.ndarray state not None,
               np.ndarray strain not None,
               np.ndarray mtx_d not None,
               CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _state, _strain, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_state, state)
    array2fmfield4(_strain, strain)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _d_biot_div(_out, coef, _state, _strain, _mtx_d, cmap.geo)
    return ret

def dw_piezo_coupling(np.ndarray out not None,
                      np.ndarray strain not None,
                      np.ndarray charge_grad not None,
                      np.ndarray mtx_g not None,
                      CMapping cmap not None,
                      int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _strain, _charge_grad, _mtx_g

    array2fmfield4(_out, out)
    array2fmfield4(_strain, strain)
    array2fmfield4(_charge_grad, charge_grad)
    array2fmfield4(_mtx_g, mtx_g)

    ret = _dw_piezo_coupling(_out, _strain, _charge_grad, _mtx_g,
                             cmap.geo, mode)
    return ret

def d_piezo_coupling(np.ndarray out not None,
                     np.ndarray strain not None,
                     np.ndarray charge_grad not None,
                     np.ndarray mtx_g not None,
                     CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _strain, _charge_grad, _mtx_g

    array2fmfield4(_out, out)
    array2fmfield4(_strain, strain)
    array2fmfield4(_charge_grad, charge_grad)
    array2fmfield4(_mtx_g, mtx_g)

    ret = _d_piezo_coupling(_out, _strain, _charge_grad, _mtx_g, cmap.geo)
    return ret

def dw_electric_source(np.ndarray out not None,
                       np.ndarray grad not None,
                       np.ndarray coef not None,
                       CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_coef, coef)

    ret = _dw_electric_source(_out, _grad, _coef, cmap.geo)
    return ret

def d_sd_diffusion(np.ndarray out not None,
                   np.ndarray grad_q not None,
                   np.ndarray grad_p not None,
                   np.ndarray grad_w not None,
                   np.ndarray div_w not None,
                   np.ndarray mtx_d not None,
                   CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _grad_q, _grad_p, _grad_w, _div_w, _mtx_d

    array2fmfield4(_out, out)
    array2fmfield4(_grad_q, grad_q)
    array2fmfield4(_grad_p, grad_p)
    array2fmfield4(_grad_w, grad_w)
    array2fmfield4(_div_w, div_w)
    array2fmfield4(_mtx_d, mtx_d)

    ret = _d_sd_diffusion(_out, _grad_q, _grad_p, _grad_w, _div_w,
                          _mtx_d, cmap.geo)
    return ret

def mulAB_integrate(np.ndarray out not None,
                    np.ndarray A not None,
                    np.ndarray B not None,
                    CMapping cmap not None,
                    mode):
    cdef int32 ret
    cdef FMField[1] _out, _A, _B

    array2fmfield4(_out, out)
    if A.ndim == 4:
        array2fmfield4(_A, A)
    else:
        array2fmfield3(_A, A)

    if B.ndim == 4:
        array2fmfield4(_B, B)
    else:
        array2fmfield3(_B, B)

    if mode == 'ATB':
        imode = 0
    elif mode == 'AB':
        imode = 1
    elif mode == 'ABT':
        imode = 2
    elif mode == 'ATBT':
        imode = 3
    else:
        imode = -1

    ret = _mulAB_integrate(_out, _A, _B, cmap.geo, imode)
    return ret

def actBfT(np.ndarray out not None,
           np.ndarray bf not None,
           np.ndarray A not None):
    cdef int32 ret
    cdef FMField[1] _out, _bf, _A

    array2fmfield4(_out, out)
    array2fmfield4(_bf, bf)
    array2fmfield4(_A, A)

    ret = _actBfT(_out, _bf, _A)
    return ret

def sym2nonsym(np.ndarray out not None,
               np.ndarray A not None):
    cdef int32 ret
    cdef FMField[1] _out, _A

    array2fmfield4(_out, out)
    array2fmfield4(_A, A)

    ret = _sym2nonsym(_out, _A)
    return ret

def dw_adj_convect1(np.ndarray out not None,
                    np.ndarray state_w not None,
                    np.ndarray grad_u not None,
                    CMapping cmap not None,
                    int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_w, _grad_u

    array2fmfield4(_out, out)
    array2fmfield4(_state_w, state_w)
    array2fmfield4(_grad_u, grad_u)

    ret = _dw_adj_convect1(_out, _state_w, _grad_u, cmap.geo, is_diff)
    return ret

def dw_adj_convect2(np.ndarray out not None,
                    np.ndarray state_w not None,
                    np.ndarray state_u not None,
                    CMapping cmap not None,
                    int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_w, _state_u

    array2fmfield4(_out, out)
    array2fmfield4(_state_w, state_w)
    array2fmfield4(_state_u, state_u)

    ret = _dw_adj_convect2(_out, _state_w, _state_u, cmap.geo, is_diff)
    return ret

def dw_st_adj_supg_c(np.ndarray out not None,
                     np.ndarray state_w not None,
                     np.ndarray state_u not None,
                     np.ndarray grad_u not None,
                     np.ndarray coef not None,
                     CMapping cmap not None,
                     np.ndarray conn not None,
                     int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_w, _state_u, _grad_u, _coef
    cdef int32 *_conn
    cdef int32 n_el, n_ep

    array2fmfield4(_out, out)
    array2fmfield1(_state_w, state_w)
    array2fmfield4(_state_u, state_u)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_coef, coef)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    ret = _dw_st_adj_supg_c(_out, _state_w, _state_u, _grad_u, _coef,
                            cmap.geo, _conn, n_el, n_ep, is_diff)
    return ret

def dw_st_adj1_supg_p(np.ndarray out not None,
                      np.ndarray state_w not None,
                      np.ndarray grad_p not None,
                      np.ndarray coef not None,
                      CMapping cmap_w not None,
                      np.ndarray conn_w not None,
                      int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _state_w, _grad_p, _coef
    cdef int32 *_conn_w
    cdef int32 n_el_w, n_ep_w

    array2fmfield4(_out, out)
    array2fmfield1(_state_w, state_w)
    array2fmfield4(_grad_p, grad_p)
    array2fmfield4(_coef, coef)
    array2pint2(&_conn_w, &n_el_w, &n_ep_w, conn_w)

    ret = _dw_st_adj1_supg_p(_out, _state_w, _grad_p, _coef,
                             cmap_w.geo, _conn_w, n_el_w, n_ep_w, is_diff)
    return ret

def dw_st_adj2_supg_p(np.ndarray out not None,
                      np.ndarray grad_u not None,
                      np.ndarray state_r not None,
                      np.ndarray coef not None,
                      CMapping cmap_u not None,
                      CMapping cmap_r not None,
                      np.ndarray conn_r not None,
                      int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _grad_u, _state_r, _coef
    cdef int32 *_conn_r
    cdef int32 n_el_r, n_ep_r

    array2fmfield4(_out, out)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield1(_state_r, state_r)
    array2fmfield4(_coef, coef)
    array2pint2(&_conn_r, &n_el_r, &n_ep_r, conn_r)

    ret = _dw_st_adj2_supg_p(_out, _grad_u, _state_r, _coef,
                             cmap_u.geo, cmap_r.geo, _conn_r, n_el_r, n_ep_r,
                             is_diff)
    return ret

def d_of_nsMinGrad(np.ndarray out not None,
                   np.ndarray grad not None,
                   np.ndarray viscosity not None,
                   CMapping cmap not None):
    cdef int32 ret
    cdef FMField[1] _out, _grad, _viscosity

    array2fmfield4(_out, out)
    array2fmfield4(_grad, grad)
    array2fmfield4(_viscosity, viscosity)

    ret = _d_of_nsMinGrad(_out, _grad, _viscosity, cmap.geo)
    return ret

def d_of_nsSurfMinDPress(np.ndarray out not None,
                         np.ndarray pressure not None,
                         float64 weight,
                         float64 bpress,
                         CMapping cmap not None,
                         int32 is_diff):
    cdef int32 ret
    cdef FMField[1] _out, _pressure

    array2fmfield4(_out, out)
    array2fmfield4(_pressure, pressure)

    ret = _d_of_nsSurfMinDPress(_out, _pressure, weight, bpress,
                                cmap.geo, is_diff)
    return ret

def d_sd_div(np.ndarray out not None,
             np.ndarray div_u not None,
             np.ndarray grad_u not None,
             np.ndarray state_p not None,
             np.ndarray div_mv not None,
             np.ndarray grad_mv not None,
             CMapping cmap_u not None,
             int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _state_p, _div_u, _grad_u
    cdef FMField[1] _div_mv, _grad_mv

    array2fmfield4(_out, out)
    array2fmfield4(_div_u, div_u)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_state_p, state_p)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)

    ret = _d_sd_div(_out, _div_u, _grad_u, _state_p, _div_mv, _grad_mv,
                    cmap_u.geo, mode)
    return ret

def d_sd_div_grad(np.ndarray out not None,
                  np.ndarray grad_u not None,
                  np.ndarray grad_w not None,
                  np.ndarray div_mv not None,
                  np.ndarray grad_mv not None,
                  np.ndarray viscosity not None,
                  CMapping cmap_u not None,
                  int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _grad_u, _grad_w, _div_mv, _grad_mv
    cdef FMField[1] _viscosity

    array2fmfield4(_out, out)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_grad_w, grad_w)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)
    array2fmfield4(_viscosity, viscosity)

    ret = _d_sd_div_grad(_out, _grad_u, _grad_w, _div_mv, _grad_mv, _viscosity,
                         cmap_u.geo, mode)
    return ret

def d_sd_convect(np.ndarray out not None,
                 np.ndarray state_u not None,
                 np.ndarray grad_u not None,
                 np.ndarray state_w not None,
                 np.ndarray div_mv not None,
                 np.ndarray grad_mv not None,
                 CMapping cmap_u not None,
                 int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _state_u, _grad_u, _state_w
    cdef FMField[1] _div_mv, _grad_mv

    array2fmfield4(_out, out)
    array2fmfield4(_state_u, state_u)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_state_w, state_w)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)

    ret = _d_sd_convect(_out, _state_u, _grad_u, _state_w, _div_mv, _grad_mv,
                        cmap_u.geo, mode)
    return ret

def d_sd_volume_dot(np.ndarray out not None,
                    np.ndarray state_p not None,
                    np.ndarray state_q not None,
                    np.ndarray div_mv not None,
                    CMapping cmap not None,
                    int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _state_p, _state_q, _div_mv

    array2fmfield4(_out, out)
    array2fmfield4(_state_p, state_p)
    array2fmfield4(_state_q, state_q)
    array2fmfield4(_div_mv, div_mv)

    ret = _d_sd_volume_dot(_out, _state_p, _state_q, _div_mv, cmap.geo, mode)
    return ret

def d_sd_st_grad_div(np.ndarray out not None,
                     np.ndarray div_u not None,
                     np.ndarray grad_u not None,
                     np.ndarray div_w not None,
                     np.ndarray grad_w not None,
                     np.ndarray div_mv not None,
                     np.ndarray grad_mv not None,
                     np.ndarray coef not None,
                     CMapping cmap_u not None,
                     int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _div_u, _grad_u, _div_w, _grad_w
    cdef FMField[1] _div_mv, _grad_mv, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_div_u, div_u)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_div_w, div_w)
    array2fmfield4(_grad_w, grad_w)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)
    array2fmfield4(_coef, coef)

    ret = _d_sd_st_grad_div(_out, _div_u, _grad_u, _div_w, _grad_w,
                            _div_mv, _grad_mv, _coef, cmap_u.geo, mode)
    return ret

def d_sd_st_supg_c(np.ndarray out not None,
                   np.ndarray state_b not None,
                   np.ndarray grad_u not None,
                   np.ndarray grad_w not None,
                   np.ndarray div_mv not None,
                   np.ndarray grad_mv not None,
                   np.ndarray coef not None,
                   CMapping cmap_u not None,
                   int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _state_b, _grad_u, _grad_w
    cdef FMField[1] _div_mv, _grad_mv, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_state_b, state_b)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_grad_w, grad_w)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)
    array2fmfield4(_coef, coef)

    ret = _d_sd_st_supg_c(_out, _state_b, _grad_u, _grad_w, _div_mv, _grad_mv,
                          _coef, cmap_u.geo, mode)
    return ret

def d_sd_st_pspg_c(np.ndarray out not None,
                   np.ndarray state_b not None,
                   np.ndarray grad_u not None,
                   np.ndarray grad_r not None,
                   np.ndarray div_mv not None,
                   np.ndarray grad_mv not None,
                   np.ndarray coef not None,
                   CMapping cmap_u not None,
                   int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _state_b, _grad_u, _grad_r
    cdef FMField[1] _div_mv, _grad_mv, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_state_b, state_b)
    array2fmfield4(_grad_u, grad_u)
    array2fmfield4(_grad_r, grad_r)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)
    array2fmfield4(_coef, coef)

    ret = _d_sd_st_pspg_c(_out, _state_b, _grad_u, _grad_r, _div_mv, _grad_mv,
                          _coef, cmap_u.geo, mode)
    return ret

def d_sd_st_pspg_p(np.ndarray out not None,
                   np.ndarray grad_r not None,
                   np.ndarray grad_p not None,
                   np.ndarray div_mv not None,
                   np.ndarray grad_mv not None,
                   np.ndarray coef not None,
                   CMapping cmap_p not None,
                   int32 mode):
    cdef int32 ret
    cdef FMField[1] _out, _grad_r, _grad_p
    cdef FMField[1] _div_mv, _grad_mv, _coef

    array2fmfield4(_out, out)
    array2fmfield4(_grad_r, grad_r)
    array2fmfield4(_grad_p, grad_p)
    array2fmfield4(_div_mv, div_mv)
    array2fmfield4(_grad_mv, grad_mv)
    array2fmfield4(_coef, coef)

    ret = _d_sd_st_pspg_p(_out, _grad_r, _grad_p, _div_mv, _grad_mv, _coef,
                          cmap_p.geo, mode)
    return ret



# -*- Mode: Python -*-
"""
Low level reference mapping functionality.
"""
import numpy as np

cdef class CMapping:

    def __cinit__(self, n_el, n_qp, dim, n_ep, mode='volume', flag=0):
        if flag:
            self.bf = np.empty((n_el, n_qp, 1, n_ep), dtype=np.float64)

        else:
            self.bf = np.empty((1, n_qp, 1, n_ep), dtype=np.float64)

        array2fmfield4(self._bf, self.bf)
        self.geo.bf = self._bf

        self.det = np.empty((n_el, n_qp, 1, 1), dtype=np.float64)
        array2fmfield4(self._det, self.det)
        self.geo.det = self._det

        self.volume = np.empty((n_el, 1, 1, 1), dtype=np.float64)
        array2fmfield4(self._volume, self.volume)
        self.geo.volume = self._volume

        if mode == 'volume':
            self.bfg = np.empty((n_el, n_qp, dim, n_ep), dtype=np.float64)
            array2fmfield4(self._bfg, self.bfg)
            self.geo.bfGM = self._bfg

            self.normal = None
            self.geo.normal = NULL

        else:
            self.bfg = None
            self.geo.bfGM = NULL

            self.normal = np.empty((n_el, n_qp, dim, 1), dtype=np.float64)
            array2fmfield4(self._normal, self.normal)
            self.geo.normal = self._normal

        self.mode = mode
        self.shape = (n_el, n_qp, dim, n_ep)

        self.geo.mode = {'volume' : MM_Volume,
                         'surface' : MM_Surface,
                         'surface_extra' : MM_SurfaceExtra}[mode]
        self.geo.nEl = self.n_el = n_el
        self.geo.nQP = self.n_qp = n_qp
        self.geo.dim = self.dim = dim
        self.geo.nEP = self.n_ep = n_ep

    def alloc_extra_data(self, n_epv):
        if self.mode != 'surface_extra':
            raise ValueError('CMapping not in surface_extra mode!')

        self.bfg = np.empty((self.n_el, self.n_qp, self.dim, n_epv),
                            dtype=np.float64)
        array2fmfield4(self._bfg, self.bfg)
        self.geo.bfGM = self._bfg

    def __str__(self):
        return 'CMapping: mode: %s, n_el %d, n_qp %d, dim: %d, n_ep: %d' \
               % ((self.mode,) + self.shape)

    def cprint(self, int32 mode=0):
        map_print(self.geo, stdout, mode)

    def describe(self,
                 np.ndarray[float64, mode='c', ndim=2] coors not None,
                 np.ndarray[int32, mode='c', ndim=2] conn not None,
                 np.ndarray[float64, mode='c', ndim=3] bfgr not None,
                 np.ndarray[float64, mode='c', ndim=4] ebfgr,
                 np.ndarray[float64, mode='c', ndim=1] weights not None):
        """
        Describe the element geometry - compute the reference element
        mapping.
        """
        cdef int32 ret = 0
        cdef FMField[1] _bfgr, _ebfgr, _weights
        cdef float64 *_coors = &coors[0, 0]
        cdef int32 n_nod = coors.shape[0]
        cdef int32 dim = coors.shape[1]
        cdef int32 *_conn = <int32 *> 0
        cdef int32 n_el = conn.shape[0]
        cdef int32 n_ep = conn.shape[1]

        if ebfgr is None:
            ebfgr = np.array([0], ndmin=4, dtype=np.float64)

        if n_el > 0:
            _conn = &conn[0, 0]

        array2fmfield3(_bfgr, bfgr)
        array2fmfield4(_ebfgr, ebfgr)
        array2fmfield1(_weights, weights)

        ret = map_describe(self.geo, _coors, n_nod, dim, _conn, n_el, n_ep,
                           _bfgr, _ebfgr, _weights)

        if ret:
            errclear()
            raise ValueError('ccore error (see above)')

    def integrate(self,
                  np.ndarray[float64, mode='c', ndim=4] out not None,
                  np.ndarray[float64, mode='c', ndim=4] arr not None,
                  int32 mode=0):
        """
        Integrate `arr` over the domain of the mapping into `out`.
        """
        cdef int32 ret = 0
        cdef FMField[1] _out, _arr

        if self.mode == 'volume':
            n_row_ok = out.shape[2] == arr.shape[2]

        else:
            n_row_ok = (((mode < 3) and (out.shape[2] == arr.shape[2]))
                        or
                        ((mode >= 3) and (out.shape[2] == 1)
                         and (arr.shape[2] == self.dim)))

        if not (((out.shape[0] == arr.shape[0]) or (arr.shape[0] == 1))
                and (out.shape[1] == 1)
                and n_row_ok
                and (out.shape[3] == arr.shape[3])
                and (out.shape[0] == self.n_el)
                and (arr.shape[1] == self.n_qp)):
            raise ValueError('incompatible shapes! (%s, out: (%d, %d, %d, %d)'
                             ', arr: (%d, %d, %d, %d))'
                             % (self.shape,
                                out.shape[0], out.shape[1],
                                out.shape[2], out.shape[3],
                                arr.shape[0], arr.shape[1],
                                arr.shape[2], arr.shape[3]))

        array2fmfield4(_out, out)
        array2fmfield4(_arr, arr)

        ret = map_integrate(self.geo, _out, _arr, mode)

        if ret:
            errclear()
            raise ValueError('ccore error (see above)')

        return ret

    def get_element_diameters(self,
                              np.ndarray[float64, mode='c', ndim=4]
                              out not None,
                              np.ndarray[int32, mode='c', ndim=2]
                              edges not None,
                              np.ndarray[float64, mode='c', ndim=2]
                              coors not None,
                              np.ndarray[int32, mode='c', ndim=2]
                              conn not None,
                              np.ndarray[int32, mode='c', ndim=1]
                              el_list not None,
                              int32 mode):
        """
        Compute diameters of selected elements.
        """
        cdef int32 ret = 0
        cdef FMField _out[1]
        cdef int32 *_edges = &edges[0, 0]
        cdef int32 edges_nr = edges.shape[0]
        cdef int32 edges_nc = edges.shape[1]
        cdef float64 *_coors = &coors[0, 0]
        cdef int32 n_nod = coors.shape[0]
        cdef int32 dim = coors.shape[1]
        cdef int32 *_conn = &conn[0, 0]
        cdef int32 n_el = conn.shape[0]
        cdef int32 n_ep = conn.shape[1]
        cdef int32 *_el_list = &el_list[0]
        cdef int32 n_el2 = el_list.shape[0]

        array2fmfield4(_out, out)

        ret = map_getElementDiameters(self.geo, _out,
                                      _edges, edges_nr, edges_nc,
                                      _coors, n_nod, dim,
                                      _conn, n_el, n_ep, _el_list, n_el2, mode)

        if ret:
            errclear()
            raise ValueError('ccore error (see above)')

        return ret

    def evaluate_bfbgm(self,
                       np.ndarray[float64, mode='c', ndim=4] bfbgr not None,
                       np.ndarray[float64, mode='c', ndim=4] ebfbgr not None,
                       np.ndarray[float64, mode='c', ndim=2] coors not None,
                       np.ndarray[int32, mode='c', ndim=2] fis not None,
                       np.ndarray[int32, mode='c', ndim=2] conn not None):
        """
        Evaluate volume base function gradients in surface quadrature
        points.
        """
        cdef int32 ret = 0
        cdef FMField[1] _bfbgr, _ebfbgr
        cdef float64 *_coors = &coors[0, 0]
        cdef int32 n_nod = coors.shape[0]
        cdef int32 dim = coors.shape[1]
        cdef int32 *_fis = &fis[0, 0]
        cdef int32 n_fa = fis.shape[0]
        cdef int32 n_fp = fis.shape[1]
        cdef int32 *_conn = &conn[0, 0]
        cdef int32 n_el = conn.shape[0]
        cdef int32 n_ep = conn.shape[1]

        if self.bfg is None:
            raise ValueError('CMapping.alloc_extra_data() not called!')

        array2fmfield4(_bfbgr, bfbgr)
        array2fmfield4(_ebfbgr, ebfbgr)

        ret = map_evaluateBFBGM(self.geo, _bfbgr, _ebfbgr, _coors, n_nod, dim,
                                _fis, n_fa, n_fp, _conn, n_el, n_ep)

        if ret:
            errclear()
            raise ValueError('ccore error (see above)')



import sys
FREECADPATH = '/usr/lib/freecad/lib/'
sys.path.append(FREECADPATH)

from FreeCAD import Base, newDocument
import Part
import Draft
import ProfileLib.RegularPolygon as Poly
import MeshPart


doc = newDocument()

radius = 0.01
height = 0.1

cyl = doc.addObject("Part::Cylinder", "cyl")
cyl.Radius = radius
cyl.Height = height

sph = doc.addObject("Part::Sphere", "sph")
sph.Radius = radius

uni = doc.addObject("Part::MultiFuse", "uni")
uni.Shapes = [cyl, sph]

ske = doc.addObject('Sketcher::SketchObject', 'Sketch')
ske.Placement = Base.Placement(Base.Vector(0, 0, 0),
                               Base.Rotation(-0.707107, 0, 0, -0.707107))
Poly.makeRegularPolygon('Sketch', 5,
                        Base.Vector(-1.2 * radius, 0.9 * height, 0),
                        Base.Vector(-0.8 * radius, 0.9 * height, 0))

cut = doc.addObject("PartDesign::Revolution", "Revolution")
cut.Sketch = ske
cut.ReferenceAxis = (ske, ['V_Axis'])
cut.Angle = 360.0

dif = doc.addObject("Part::Cut", "dif")
dif.Base = uni
dif.Tool = cut

cyl1 = doc.addObject("Part::Cylinder", "cyl1")
cyl1.Radius = 0.2 * radius
cyl1.Height = 1.1 * height
cyl1.Placement = Base.Placement(Base.Vector(-1.1 * radius, 0, -0.2 * height),
                                Base.Rotation(0, 0, 0, 1))

arr = Draft.makeArray(cyl1, Base.Vector(1, 0, 0), Base.Vector(0, 1, 0), 2, 2)
arr.ArrayType = "polar"
arr.NumberPolar = 6

dif2 = doc.addObject("Part::Cut", "dif2")
dif2.Base = dif
dif2.Tool = arr

cyl2 = doc.addObject("Part::Cylinder", "cyl2")
cyl2.Radius = 0.3 * radius
cyl2.Height = height

dif3 = doc.addObject("Part::Cut", "dif3")
dif3.Base = dif2
dif3.Tool = cyl2

cyl3 = doc.addObject("Part::Cylinder", "cyl3")
cyl3.Radius = 0.3 * radius
cyl3.Height = height
cyl3.Placement = Base.Placement(Base.Vector(0, 0, height),
                                Base.Rotation(0, 0, 0, 1))

tip1 = doc.addObject("Part::Box", "tip1")
tip1.Length = radius
tip1.Width = 2 * radius
tip1.Height = 3 * radius
tip1.Placement = Base.Placement(Base.Vector(0, -radius, 1.71 * height),
                                Base.Rotation(Base.Vector(0, 1, 0), -10),
                                Base.Vector(0, 0, 3 * radius))

tip2 = doc.addObject("Part::Mirroring", "tip2")
tip2.Source = tip1
tip2.Normal = (1, 0, 0)

tip3 = doc.addObject("Part::MultiFuse", "tip3")
tip3.Shapes = [tip1, tip2]

dif4 = doc.addObject("Part::Cut", "dif4")
dif4.Base = cyl3
dif4.Tool = tip3

uni2 = doc.addObject("Part::MultiFuse", "uni2")
uni2.Shapes = [cyl2, dif4]

doc.recompute()

Part.export([dif3, uni2], 'screwdriver_full.step')
doc.saveAs('screwdriver_full.FCStd')



# -*- Mode: Python -*-
"""
Low level functions.
"""
cimport cython

cimport sfepy.discrete.common.extmods._geommech as gmch
from types cimport int32

from sfepy.discrete.common.extmods._fmfield cimport\
    (FMField, FMF_SetCell, array2fmfield4)

cimport numpy as np
import numpy as np

def geme_mulAVSB3py(np.ndarray vs not None,
                   np.ndarray inp not None):
    cdef int32 ret
    cdef FMField[1] _out, _vs, _inp

    out = np.zeros_like(inp)
    array2fmfield4(_out, out)
    array2fmfield4(_vs, vs)
    array2fmfield4(_inp, inp)

    ret = 0
    for ii in range(out.shape[0]):
        FMF_SetCell(_out, ii);
        FMF_SetCell(_vs, ii);
        FMF_SetCell(_inp, ii);
        _ret = gmch.geme_mulAVSB3(_out, _vs, _inp)
        ret = ret and _ret

    return out, ret



#!/usr/bin/env python
"""
Generate simplex legendre 2D basis coffecients and exponents matrices and
save them to legendre2D_simplex_coefs.txt and legendre2D_simplex_expos.txt
"""
from argparse import ArgumentParser
from sympy import symbols
from sympy import jacobi_poly as jacobi_P
from sympy import expand_mul as expand
from sympy import cancel, simplify

import numpy as np

from sfepy.base.ioutils import InDir
from sfepy.discrete.dg.poly_spaces import iter_by_order, get_n_el_nod

helps = {
    'max_order' :
        'maximum order of polynomials [default: %(default)s]',
    'output_dir':
        'output directory',
}


def main():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('-m', '--max-order', metavar='order', type=int,
                        action='store', dest='max_order',
                        default=10, help=helps['max_order'])
    parser.add_argument('output_dir', help=helps['output_dir'])
    options = parser.parse_args()

    indir = InDir(options.output_dir)

    a, b = symbols("a, b")
    r, s = symbols("r, s")
    x, y = symbols("x, y")
    order = options.max_order
    dim = 2
    n_el_nod = get_n_el_nod(order, dim)  # number of DOFs per element

    simplexP = []
    exponentM = np.zeros((n_el_nod, 3), dtype=np.int32)
    coefM = np.zeros((n_el_nod, n_el_nod))
    exponentList = []
    for m, idx in enumerate(iter_by_order(order, dim)):
        # print(m, idx)
        exponentM[m, :dim] = idx
        pa = jacobi_P(idx[0], 0, 0, a)
        pb = jacobi_P(idx[1], 2 * idx[0] + 1, 0, b) * (1 - b) ** idx[0]
        # print("P_{} = {}".format(m, pa*pb))
        polrs = cancel((pa * pb).subs(b, s).subs(
                a, 2 * (1 + r) / (1 - s) - 1))
        # print("P_{} = {}".format(m, polrs))
        polxy = expand(polrs.subs(r, 2 * x - 1).subs(s, 2 * y - 1))
        # polxy = expand(polrs.subs(r, x).subs(s, y))

        simplexP.append(simplify(polxy))
        exponentList.append(x ** idx[0] * y ** idx[1])
        for j, exponent in enumerate(exponentList):
            coefM[m, j] = simplexP[m].as_coefficients_dict()[exponent]
        print("P_{}{} = {}".format(m, idx, simplexP[m]))
        print()

    np.savetxt(indir("legendre2D_simplex_expos.txt"), exponentM, fmt="%d")
    # are coefs always integers?
    np.savetxt(indir("legendre2D_simplex_coefs.txt"), coefM, fmt="%d")


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
This is a script for quick VTK-based visualizations of finite element
computations results.

Examples
--------
  The examples assume that run_tests.py has been run successfully and the
  resulting data files are present.

  - View data in output-tests/test_navier_stokes.vtk.

    $ python resview.py output-tests/test_navier_stokes.vtk

  - Customize the above output:
    plot0: field "p", switch on edges,
    plot1: field "u", surface with opacity 0.4, glyphs scaled by factor 2e-2.

    $ python resview.py output-tests/test_navier_stokes.vtk -f p:e:p0 u:o.4:p1 u:g:f2e-2:p1

  - As above, but glyphs are scaled by the factor determined automatically as
    20% of the minimum bounding box size.

    $ python resview.py output-tests/test_navier_stokes.vtk -f p:e:p0 u:o.4:p1 u:g:f10%:p1

  - View data and take a screenshot.

    $ python resview.py output-tests/test_poisson.vtk -o image.png

  - Take a screenshot without a window popping up.

    $ python resview.py output-tests/test_poisson.vtk -o image.png --off-screen

  - Create animation from output-tests/test_time_poisson.*.vtk.

    $ python resview.py output-tests/test_time_poisson.*.vtk -a mov.mp4

  - Create animation from output-tests/test_hyperelastic.*.vtk,
    set frame rate to 3, plot displacements and mooney_rivlin_stress.

    $ python resview.py output-tests/test_hyperelastic_TL.*.vtk -f u:wu:e:p0 mooney_rivlin_stress:p1 -a mov.mp4 -r 3
"""
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from ast import literal_eval
import numpy as nm
import os.path as osp

import pyvista as pv
from vtk.util.numpy_support import numpy_to_vtk

cache = {}


def get_camera_position(bounds, azimuth, elevation, distance=None, zoom=1.):
    phi, psi = nm.deg2rad(azimuth), nm.deg2rad(elevation)
    bounds = nm.asarray(bounds)

    if distance is not None:
        r = distance / zoom
    else:
        r = max(bounds[1::2] - bounds[::2]) * 2.0 / zoom

    center = (bounds[1::2] + bounds[::2]) * 0.5

    # camera position
    position = (r * nm.cos(phi) * nm.sin(psi),
                r * nm.sin(phi) * nm.sin(psi),
                r * nm.cos(psi))

    # view up
    view_up = (0, 0, 1)
    if abs(elevation) < 5. or abs(elevation) > 175.:
        view_up = (nm.sin(phi), nm.cos(phi), 0)

    return [position, tuple(center), view_up]


def parse_options(opts, separator=':'):
    out = {}
    if opts is None:
        return out

    for v in opts.split(separator):
        if len(v) < 2:
            val = True
        elif v[1:].isalpha():
            val = v[1:]
        elif v[-1] == '%':
            val = ('%', float(v[1:-1]))
        else:
            try:
                val = literal_eval(v[1:])
            except ValueError:
                val = v[1:]

        out[v[0]] = val

    return out


def make_cells_from_conn(conns, convert_to_vtk_type):
    cells, cell_type, offset = [], [], []
    _offset = 0
    for ctype, conn in conns.items():
        nc, np = conn.shape

        aux = nm.empty((nc, np + 1), dtype=int)
        aux[:, 0] = np
        aux[:, 1:] = conn
        cells.append(aux.ravel())

        cell_type.append(nm.full(nc, convert_to_vtk_type[ctype]))
        offset.append(nm.arange(nc) * (np + 1) + _offset)
        _offset += nc

    cells = nm.concatenate(cells)
    cell_type = nm.concatenate(cell_type)
    offset = nm.concatenate(offset)

    return cells, cell_type, offset


def add_mat_id_to_grid(grid, cell_groups):
    val = numpy_to_vtk(cell_groups)
    val.SetName('mat_id')
    grid.GetCellData().AddArray(val)
    return grid


def read_mesh(filenames, step=None, print_info=True, ret_n_steps=False,
              use_cache=True):
    _, ext = osp.splitext(filenames[0])
    if ext in ['.vtk', '.vtu']:
        fstep = 0 if step is None else step
        fname = filenames[fstep]
        key = (fname, fstep)
        if key not in cache or not use_cache:
            cache[key] = pv.UnstructuredGrid(fname)
        mesh = cache[key]
        cache['n_steps'] = len(filenames)
    elif ext in ['.xdmf', '.xdmf3']:
        import meshio
        try:
            from meshio._common import meshio_to_vtk_type

        except ImportError:
            from meshio._vtk_common import meshio_to_vtk_type

        fname = filenames[0]
        key = (fname, step)
        if key not in cache:
            reader = meshio.xdmf.TimeSeriesReader(fname)
            points, _cells = reader.read_points_cells()
            points = nm.asarray(points)
            if points.shape[1] < 3:
                points = nm.pad(points, [(0, 0), (0, 3 - points.shape[1])])
            _dcells = {ct.type: ct.data for ct in _cells}

            cells, cell_type, offset = make_cells_from_conn(
                _dcells, meshio_to_vtk_type,
            )

            if not reader.num_steps:
                grid = pv.UnstructuredGrid(offset, cells, cell_type, points)
                add_mat_id_to_grid(grid, mesh.cmesh.cell_groups)
                cache[(fname, 0)] = grid

            grids = {}
            time = []
            for _step in range(reader.num_steps):
                grid = pv.UnstructuredGrid(offset, cells, cell_type, points)
                t, pd, cd = reader.read_data(_step)
                for dk, dv in pd.items():
                    val = numpy_to_vtk(dv)
                    val.SetName(dk)
                    grid.GetPointData().AddArray(val)

                for dk, dv in cd.items():
                    val = numpy_to_vtk(nm.vstack(dv).squeeze())
                    val.SetName(dk)
                    grid.GetCellData().AddArray(val)

                grids[t] = grid
                time.append(t)

            time.sort()
            for _step, t in enumerate(time):
                cache[(fname, _step)] = grids[t]

            cache[(fname, None)] = cache[(fname, 0)]
            cache['n_steps'] = reader.num_steps

        mesh = cache[key]

    elif ext in ['.h5', '.h5x']:
        vtk_cell_types = {'1_1': 1, '1_2': 3, '2_2': 3, '3_2': 3,
                          '2_3': 5, '2_4': 9, '3_4': 10, '3_8': 12}
        # Custom sfepy format.
        fname = filenames[0]
        key = (fname, step)
        if key not in cache:
            from sfepy.discrete.fem.meshio import MeshIO

            io = MeshIO.any_from_filename(fname)

            mesh = io.read()
            desc = mesh.descs[0]
            nv, dim = mesh.coors.shape

            points = nm.c_[mesh.coors, nm.zeros((nv, 3 - dim))]
            cells, cell_type, offset = make_cells_from_conn(
                {desc: mesh.get_conn(desc)}, vtk_cell_types,
            )

            steps, times, nts = io.read_times()
            if not len(steps):
                grid = pv.UnstructuredGrid(offset, cells, cell_type, points)
                add_mat_id_to_grid(grid, mesh.cmesh.cell_groups)
                cache[(fname, 0)] = grid

            for ii, _step in enumerate(steps):
                grid = pv.UnstructuredGrid(offset, cells, cell_type, points)
                datas = io.read_data(_step)
                for dk, data in datas.items():
                    vval = data.data
                    if 1 < len(data.dofs) < 3:
                        vval = nm.c_[vval,
                                     nm.zeros((len(vval), 3 - len(data.dofs)))]

                    if data.mode == 'vertex':
                        val = numpy_to_vtk(vval)
                        val.SetName(dk)
                        grid.GetPointData().AddArray(val)

                    else:
                        val = numpy_to_vtk(vval[:, 0, :, 0])
                        val.SetName(dk)
                        grid.GetCellData().AddArray(val)

                cache[(fname, ii)] = grid

            cache[(fname, None)] = cache[(fname, 0)]
            cache['n_steps'] = len(steps)

        mesh = cache[key]

    else:
        raise ValueError('unknown file format! (%s)' % ext)

    if print_info:
        arrs = {'s': [], 'v': [], 'o': []}
        for aname in mesh.array_names:
            if len(mesh[aname].shape) == 1 or mesh[aname].shape[1] == 1:
                arrs['s'].append(aname)
            elif mesh[aname].shape[1] == 3:
                arrs['v'].append(aname)
            else:
                arrs['o'].append(aname + '(%d)' % mesh[aname].shape[1])

        step_info = ' (step %d)' % step if step else ''
        print('mesh from %s%s:' % (fname, step_info))
        print('  points:  %d' % mesh.n_points)
        print('  cells:   %d' % mesh.n_cells)
        print('  bounds:  %s' % list(zip(nm.min(mesh.points, axis=0),
                                         nm.max(mesh.points, axis=0))))
        if len(arrs['s']) > 0:
            print('  scalars: %s' % ', '.join(arrs['s']))
        if len(arrs['v']) > 0:
            print('  vectors: %s' % ', '.join(arrs['v']))
        if len(arrs['o']) > 0:
            print('  others:  %s' % ', '.join(arrs['o']))
        print('  steps:   %d' % cache['n_steps'])

    if ret_n_steps:
        return mesh, cache['n_steps']
    else:
        return mesh


def pv_plot(filenames, options, plotter=None, step=None,
            scalar_bar_limits=None, ret_scalar_bar_limits=False,
            step_inc=None, use_cache=True):
    plots = {}
    color = None

    if plotter is None:
        plotter = pv.Plotter()

    fstep = (step if step is not None else options.step)
    if step_inc is not None:
        plotter.clear()
        fstep += step_inc
    if fstep < 0:
        fstep = 0
    if hasattr(plotter, 'resview_n_steps'):
        if fstep >= plotter.resview_n_steps:
            fstep = plotter.resview_n_steps - 1

    mesh, n_steps = read_mesh(filenames, fstep, ret_n_steps=True,
                              use_cache=use_cache)
    steps = {fstep: mesh}

    bbox_sizes = nm.diff(nm.reshape(mesh.bounds, (-1, 2)), axis=1)
    ii = nm.where(bbox_sizes > 0)[0]
    tdim = len(ii)
    if options.position_vector is None:
        if len(ii):
            ipv = ii[-1]

        else:
            ipv = 2
            print('WARNING: zero size mesh!')

        options.position_vector = [0, 0, 0]
        options.position_vector[ipv] = 1.6

    plotter.resview_step, plotter.resview_n_steps = fstep, n_steps

    fields_map = {}
    if len(options.fields_map) > 0:
        for cg, fields in options.fields_map:
            for field in fields.split(','):
                fields_map[field.strip()] = int(cg)

    if len(options.fields) == 0:
        fields = []
        position = 0
        for field in steps[fstep].array_names:
            if field in ['node_groups', 'mat_id']:
                continue

            fval = steps[fstep][field]
            bnds = steps[fstep].bounds
            mesh_size = (nm.array(bnds[1::2]) - nm.array(bnds[::2])).max()
            is_vector_field = len(fval.shape) > 1
            is_point_field = fval.shape[0] == steps[fstep].n_points
            if is_vector_field and is_point_field:
                scale = mesh_size * 0.15 / nm.linalg.norm(fval, axis=1).max()
                fields.append((field, 'vw:p%d' % position))
                fields.append((field, 'vs:o.4:p%d' % position))
                fields.append((field, 'g:f%e:p%d' % (scale, position)))
            else:
                fields.append((field, 'p%d' % position))

            position += 1

        if len(fields) == 0:
            fields.append(('mat_id', 'p0'))
    else:
        fields = options.fields

    plot_id = 0

    scalar_bars = {}
    for field, fopts in fields:
        opts = parse_options(fopts)
        plot_info = []

        if field == '0':
            field = None
            color = 'white'

        if field == '1':
            field = None
            color = 'black'

        if 's' in opts and step is None:  # plot data from a given step
            fstep = opts['s']

        if fstep not in steps:
            steps[fstep] = read_mesh(filenames, step=fstep,
                                     use_cache=use_cache)

        pipe = [steps[fstep].copy()]

        if field in fields_map:  # subregion
            mat_val = fields_map[field]
        elif 'm' in opts:
            mat_val = opts['m']
        else:
            mat_val = None

        if mat_val:
            if isinstance(mat_val, int):
                mat_val = [mat_val, mat_val]

            pipe.append(pipe[-1].threshold(value=mat_val,
                        scalars='mat_id', preference='cell'))

        if 'r' in opts:  # recalculate cell data to point data
            pipe.append(pipe[-1].cell_data_to_point_data())

        opacity = opts.get('o', options.opacity)  # mesh opacity
        show_edges = opts.get('e', options.show_edges)  # edge visibility
        style = {'s': 'surface',
                 'w': 'wireframe',
                 'p': 'points'}[opts.get('v', 's')]  # set style

        warp = opts.get('w', options.warp)  # warp mesh
        factor = opts.get('f', options.factor)
        if isinstance(factor, tuple):
            ws = nm.diff(nm.reshape(pipe[-1].bounds, (-1, 2)), axis=1)
            size = ws[ws > 0.0].min()
            fmax = nm.abs(pipe[-1][field]).max()
            factor = 0.01 * float(factor[1]) * size / fmax

        if warp:
            field_data = pipe[-1][warp]
            if field_data.ndim == 1:
                field_data.shape = (-1, 1)
            nc = field_data.shape[1]
            if nc == 1:  # Warp by scalar.
                pipe.append(pipe[-1].copy())
                pipe[-1].points[:, 2] += field_data[:, 0] * factor

            elif nc == 3:
                pipe.append(pipe[-1].copy())
                pipe[-1].points += field_data * factor

            else:
                raise ValueError('warp mesh: scalar or vector field required!')

            plot_info.append('warp=%s, factor=%.2e' % (warp, factor))

        position = opts.get('p', 0)  # determine plotting slot
        bnds = pipe[-1].bounds
        if 'p' in opts:
            size = nm.array(bnds[1::2]) - nm.array(bnds[::2])
            pipe.append(pipe[-1].copy())
            shift = position * size * nm.array(options.position_vector)
            pipe[-1].translate(shift)

        if opts.get('l', options.outline):  # outline
            plotter.add_mesh(pipe[-1].outline(), color='k')

        scalar = field
        scalar_label = scalar
        is_vector_field = field is not None and len(pipe[-1][field].shape) > 1
        is_point_field = (field is not None and
                          pipe[-1][field].shape[0] == pipe[-1].n_points)
        if is_vector_field:
            field_data = pipe[-1][field]
            scalar = field + '_magnitude'
            scalar_label = f'|{field}|'
            pipe[-1][scalar] = nm.linalg.norm(field_data, axis=1)

        if 'g' in opts and is_vector_field and is_point_field:  # glyphs
            pipe[-1][field] *= factor
            pipe[-1].set_active_vectors(field)
            pipe.append(pipe[-1].arrows)
            style = ''
            plot_info.append('glyphs=%s, factor=%.2e' % (field, factor))
        elif 'c' in opts and is_vector_field:  # select field component
            comp = opts['c']
            scalar = field + '_%d' % comp
            pipe[-1][scalar] = field_data[:, comp]
        elif 't' in opts:  # streamlines
            npts = opts.get('t')
            if npts is True:
                npts = 20

            if is_vector_field:
                sl_vector = field
                sl_pipe = pipe[-1]
            else:
                sl_vector = 'gradient'
                sl_pipe = pipe[-1].compute_derivative(scalars=field)

            cmin, cmax = sl_pipe.bounds[::2], sl_pipe.bounds[1::2]
            if tdim == 2:
                streamlines = sl_pipe.streamlines(vectors=sl_vector,
                                                  pointa=cmin, pointb=cmax,
                                                  n_points=npts,
                                                  max_time=1e12)

            else:
                radius = 0.5 * nm.linalg.norm(nm.array(cmax) - nm.array(cmin))
                streamlines = sl_pipe.streamlines(vectors=sl_vector,
                                                  source_radius=radius,
                                                  n_points=npts,
                                                  max_time=1e12)

            pipe.append(streamlines)

        plotter.add_mesh(pipe[-1], scalars=scalar, color=color,
                         style=style, show_edges=show_edges,
                         opacity=opacity,
                         cmap=options.color_map,
                         show_scalar_bar=False, label=scalar_label)

        bnds = pipe[-1].bounds
        if position not in plots:
            plots[position] = []

        plot_info = ':' + ','.join(plot_info) if len(plot_info) > 0 else ''
        plot_info = '%s(step %d)%s' % (scalar, fstep, plot_info)
        plots[position].append(((bnds[::2], bnds[1::2]), plot_info))

        if options.show_scalar_bars and scalar:
            if scalar not in scalar_bars:
                scalar_bars[scalar_label] = []

            field_data = pipe[-1][scalar]
            limits = (nm.min(field_data), nm.max(field_data))
            scalar_bars[scalar_label].append((limits, plotter.mapper,
                                              position))

        plot_id += 1

    if options.show_scalar_bars:
        if scalar_bar_limits is None:
            scalar_bar_limits = {}
            for k, vs in scalar_bars.items():
                limits = (nm.min([v[0] for v, _, _ in vs]),
                          nm.max([v[1] for v, _, _ in vs]))
                scalar_bar_limits[k] = limits

        width, height = options.scalar_bar_size
        position_x, position_y, shift_x, shift_y = options.scalar_bar_position
        nslots = len(scalar_bars)
        for k, vs in scalar_bars.items():
            clim = scalar_bar_limits[k]
            for _, mapper, _ in vs:
                mapper.scalar_range = clim
            _, mapper, slot = vs[0]

            slot_x = (nslots - slot - 1) if shift_x < 0 else slot
            x_pos = position_x + slot_x * width * shift_x
            slot_y = (nslots - slot - 1) if shift_y < 0 else slot
            y_pos = position_y + slot_y * height * shift_y

            plotter.add_scalar_bar(title=k,
                                   position_x=x_pos, position_y=y_pos,
                                   width=width, height=height,
                                   n_labels=2, mapper=mapper)

    if options.show_labels and len(plots) > 1:
        labels, points = [], []
        for k, v in plots.items():
            bnds = (nm.min(nm.array([iv[0][0] for iv in v]), axis=0),
                    nm.max(nm.array([iv[0][1] for iv in v]), axis=0))
            labels.append('plot:%d' % k)
            size = bnds[1] - bnds[0]
            olpos = options.label_position
            points.append(bnds[0] + nm.array(olpos[:3]) * size * olpos[3])

        plotter.add_point_labels(nm.array(points), labels)

    for k, v in plots.items():
        print('plot %d: %s' % (k, '; '.join(iv[1] for iv in v)))

    if ret_scalar_bar_limits:
        return plotter, scalar_bar_limits
    else:
        return plotter


class OptsToListAction(Action):
    separator = '='

    def __call__(self, parser, namespace, values, option_string=None):
        out = []
        for item in values:
            s = item.split(self.separator, 1)
            out.append((s[0].strip(), s[1].strip() if len(s) > 1 else None))

        setattr(namespace, self.dest, out)


class FieldOptsToListAction(OptsToListAction):
    separator = ':'


class StoreNumberAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, literal_eval(values))


helps = {
    'fields':
        'fields to plot, options separated by ":" are possible:\n'
        '"cX" - plot only Xth field component; '
        '"e" - print edges; '
        '"fX" - scale factor for warp/glyphs, see --factor option; '
        '"g - glyphs (for vector fields only), scale by factor; '
        '"tX" - plot X streamlines, gradient employed for scalar fields; '
        '"mX" - plot cells with mat_id=X; '
        '"oX" - set opacity to X; '
        '"pX" - plot in slot X; '
        '"r" - recalculate cell data to point data; '
        '"sX" - plot data in step X; '
        '"vX" - plotting style: s=surface, w=wireframe, p=points; '
        '"wX" - warp mesh by vector field X, scale by factor',
    'fields_map':
        'map fields and cell groups, e.g. 1:u1,p1 2:u2,p2',
    'outline':
        'plot mesh outline',
    'warp':
        'warp mesh by vector field',
    'factor':
        'scaling factor for mesh warp and glyphs.'
        ' Append "%%" to scale relatively to the minimum bounding box size.',
    'edges':
        'plot cell edges',
    'opacity':
        'set opacity [default: %(default)s]',
    'color_map':
        'set color_map, e.g. hot, cool, bone, etc. [default: %(default)s]',
    'axes_options':
        'options for directional axes, e.g. xlabel="z1" ylabel="z2",'
        ' zlabel="z3"',
    'no_axes':
        'hide orientation axes',
    'no_scalar_bars':
        'hide scalar bars',
    'position_vector':
        'define positions of plots [default: "0, 0, 1.6"]',
    'view':
        'camera azimuth, elevation angles, and optionally zoom factor'
        ' [default: "225,75,0.9"]',
    'animation':
        'create animation, mp4 file type supported',
    'framerate':
        'set framerate for animation',
    'screenshot':
        'save screenshot to file',
    'off_screen':
        'off screen plots, e.g. when screenshotting',
    'no_labels':
        'hide plot labels',
    'label_position':
        'define position of plot labels [default: "-1, -1, 0, 0.2"]',
    'scalar_bar_size':
        'define size of scalar bars [default: "0.15, 0.05"]',
    'scalar_bar_position':
        'define position of scalar bars [default: "0.8, 0.02, 0, 1.5"]',
    'step':
        'select data in a given time step',
    '2d_view':
        '2d view of XY plane',
}


def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-f', '--fields', metavar='field_spec',
                        action=FieldOptsToListAction, nargs="+", dest='fields',
                        default=[], help=helps['fields'])
    parser.add_argument('--fields-map', metavar='map',
                        action=FieldOptsToListAction, nargs="+",
                        dest='fields_map',
                        default=[], help=helps['fields_map'])
    parser.add_argument('-s', '--step', metavar='step',
                        action=StoreNumberAction, dest='step',
                        default=0, help=helps['step'])
    parser.add_argument('-l', '--outline',
                        action='store_true', dest='outline',
                        default=False, help=helps['outline'])
    parser.add_argument('-e', '--edges',
                        action='store_true', dest='show_edges',
                        default=False, help=helps['edges'])
    parser.add_argument('-w', '--warp', metavar='field',
                        action='store', dest='warp',
                        default=None, help=helps['warp'])
    parser.add_argument('--factor', metavar='factor',
                        action=StoreNumberAction, dest='factor',
                        default=1., help=helps['factor'])
    parser.add_argument('--opacity', metavar='opacity',
                        action=StoreNumberAction, dest='opacity',
                        default=1., help=helps['opacity'])
    parser.add_argument('--color-map', metavar='cmap',
                        action='store', dest='color_map',
                        default='viridis', help=helps['color_map'])
    parser.add_argument('--axes-options', metavar='options',
                        action=OptsToListAction, nargs="+",
                        dest='axes_options',
                        default=[], help=helps['axes_options'])
    parser.add_argument('--no-axes',
                        action='store_false', dest='axes_visibility',
                        default=True, help=helps['no_axes'])
    parser.add_argument('--position-vector', metavar='position_vector',
                        action=StoreNumberAction, dest='position_vector',
                        default=None, help=helps['position_vector'])
    parser.add_argument('--no-labels',
                        action='store_false', dest='show_labels',
                        default=True, help=helps['no_labels'])
    parser.add_argument('--label-position', metavar='position',
                        action=StoreNumberAction, dest='label_position',
                        default=[-1, -1, 0, 0.2], help=helps['label_position'])
    parser.add_argument('--no-scalar-bars',
                        action='store_false', dest='show_scalar_bars',
                        default=True, help=helps['no_scalar_bars'])
    parser.add_argument('--scalar-bar-size', metavar='size',
                        action=StoreNumberAction, dest='scalar_bar_size',
                        default=[0.15, 0.05],
                        help=helps['scalar_bar_size'])
    parser.add_argument('--scalar-bar-position', metavar='position',
                        action=StoreNumberAction, dest='scalar_bar_position',
                        default=[0.8, 0.02, 0, 1.5],
                        help=helps['scalar_bar_position'])
    parser.add_argument('-v', '--view', metavar='position',
                        action=StoreNumberAction, dest='camera',
                        default=[225, 75, 0.9], help=helps['view'])
    parser.add_argument('-a', '--animation', metavar='output_file',
                        action='store', dest='anim_output_file',
                        default=None, help=helps['animation'])
    parser.add_argument('-r', '--frame-rate', metavar='rate',
                        action=StoreNumberAction, dest='framerate',
                        default=2.5, help=helps['framerate'])
    parser.add_argument('-o', '--screenshot', metavar='output_file',
                        action='store', dest='screenshot',
                        default=None, help=helps['screenshot'])
    parser.add_argument('--off-screen',
                        action='store_true', dest='off_screen',
                        default=False, help=helps['off_screen'])
    parser.add_argument('-2', '--2d-view',
                        action='store_true', dest='view_2d',
                        default=False, help=helps['2d_view'])

    parser.add_argument('filenames', nargs='+')
    options = parser.parse_args()

    pv.set_plot_theme("document")
    plotter = pv.Plotter(off_screen=options.off_screen)

    if options.anim_output_file:
        _, n_steps = read_mesh(options.filenames, ret_n_steps=True)
        # dry run
        scalar_bar_limits = None
        if options.axes_visibility:
            plotter.add_axes(**dict(options.axes_options))
        for step in range(n_steps):
            plotter, sb_limits = pv_plot(options.filenames, options,
                                         plotter=plotter, step=step,
                                         ret_scalar_bar_limits=True)
            if scalar_bar_limits is None:
                scalar_bar_limits = {k: [] for k in sb_limits.keys()}

            for k, v in sb_limits.items():
                scalar_bar_limits[k].append(v)

        if options.camera:
            zoom = options.camera[2] if len(options.camera) > 2 else 1.
            cpos = get_camera_position(plotter.bounds,
                                       options.camera[0], options.camera[1],
                                       zoom=zoom)
            plotter.set_position(cpos[0])
            plotter.set_focus(cpos[1])
            plotter.set_viewup(cpos[2])

        anim_filename = options.anim_output_file
        plotter.open_movie(anim_filename, options.framerate)
        plotter.show(auto_close=False)

        for k in scalar_bar_limits.keys():
            lims = scalar_bar_limits[k]
            clim = (nm.min([v[0] for v in lims]),
                    nm.max([v[1] for v in lims]))
            scalar_bar_limits[k] = clim

        # plot frames
        for step in range(n_steps):
            plotter.clear()
            plotter = pv_plot(options.filenames, options, plotter=plotter,
                              step=step, scalar_bar_limits=scalar_bar_limits)
            if options.axes_visibility:
                plotter.add_axes(**dict(options.axes_options))

            plotter.write_frame()

        plotter.close()
    else:
        plotter = pv_plot(options.filenames, options, plotter=plotter)
        if options.axes_visibility:
            plotter.add_axes(**dict(options.axes_options))

        plotter.add_key_event(
            'Prior', lambda: pv_plot(options.filenames,
                                     options,
                                     step=plotter.resview_step,
                                     step_inc=-1,
                                     plotter=plotter)
        )
        plotter.add_key_event(
            'Next', lambda: pv_plot(options.filenames,
                                    options,
                                    step=plotter.resview_step,
                                    step_inc=1,
                                    plotter=plotter)
        )

        if options.view_2d:
            plotter.view_xy()
            plotter.show(screenshot=options.screenshot)
        else:
            if options.camera:
                zoom = options.camera[2] if len(options.camera) > 2 else 1.
                cpos = get_camera_position(plotter.bounds,
                                           options.camera[0],
                                           options.camera[1],
                                           zoom=zoom)
            else:
                cpos = None

            plotter.show(cpos=cpos, screenshot=options.screenshot)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# 06.06.2017, c
"""
Solve a coupled two-scale problem in parallel. One computational node is solving
a macroscopic equation while the others are solving local microscopic problems
and homogenized coefficients.

Run this script as::

    mpiexec -n 4 simple_homog_mpi.py examples/homogenization/nonlinear_hyperelastic_mM.py

"""
from __future__ import print_function
from __future__ import absolute_import
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from sfepy.base.base import output, Struct
from sfepy.base.conf import ProblemConf, get_standard_keywords
from sfepy.applications import PDESolverApp
from sfepy.homogenization.homogen_app import HomogenizationApp
import simple

import sfepy.base.multiproc_mpi as multi_mpi


helps = {k: simple.helps[k] for k in ['debug', 'conf', 'options', 'define',
                                      'filename', 'output_format', 'log',
                                      'quiet']}
helps.update({
    'debug_mpi': 'log MPI communication',
})


def main():
    # if multi_mpi.cpu_count() < 2:
    #     raise ValueError('MPI mode - the number of nodes is less then 2!')

    if multi_mpi.mpi_rank == multi_mpi.mpi_master:
        # MPI master node - solve problem at macro scale
        parser = ArgumentParser(description=__doc__,
                                formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument('--debug',
                            action='store_true', dest='debug',
                            default=False, help=helps['debug'])
        parser.add_argument('--debug_mpi',
                            action='store_true', dest='debug_mpi',
                            default=False, help=helps['debug_mpi'])
        parser.add_argument('-c', '--conf', metavar='"key : value, ..."',
                            action='store', dest='conf', type=str,
                            default=None, help=helps['conf'])
        parser.add_argument('-O', '--options', metavar='"key : value, ..."',
                            action='store', dest='app_options', type=str,
                            default=None, help=helps['options'])
        parser.add_argument('-d', '--define', metavar='"key : value, ..."',
                            action='store', dest='define_args', type=str,
                            default=None, help=helps['define'])
        parser.add_argument('-o', metavar='filename',
                            action='store', dest='output_filename_trunk',
                            default=None, help=helps['filename'])
        parser.add_argument('--format', metavar='format',
                            action='store', dest='output_format',
                            default=None, help=helps['output_format'])
        parser.add_argument('--log', metavar='file',
                            action='store', dest='log',
                            default=None, help=helps['log'])
        parser.add_argument('-q', '--quiet',
                            action='store_true', dest='quiet',
                            default=False, help=helps['quiet'])
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('filename_in', nargs='?')
        options = parser.parse_args()

        for k in ['save_ebc', 'save_ebc_nodes', 'save_regions',
                  'save_regions_as_groups', 'save_field_meshes', 'solve_not']:
            setattr(options, k, False)

        if options.debug:
            from sfepy.base.base import debug_on_error; debug_on_error()

        if options.debug_mpi:
            multi_mpi.set_logging_level('debug')

        filename_in = options.filename_in
        output.set_output(filename=options.log,
                          quiet=options.quiet,
                          combined=options.log is not None)

        required, other = get_standard_keywords()
        conf = ProblemConf.from_file_and_options(filename_in, options,
            required, other, define_args=options.define_args)

        opts = conf.options
        nslaves = multi_mpi.cpu_count() - 1
        opts.n_mpi_homog_slaves = nslaves
        output_prefix = opts.get('output_prefix', 'sfepy:')

        app = PDESolverApp(conf, options, output_prefix)
        if hasattr(opts, 'parametric_hook'):  # Parametric study.
            parametric_hook = conf.get_function(opts.parametric_hook)
            app.parametrize(parametric_hook)
        app()

        multi_mpi.master_send_task('finalize', None)
    else:
        # MPI slave mode - calculate homogenized coefficients
        homogen_app = None
        done = False
        rank = multi_mpi.mpi_rank
        while not done:
            task, data = multi_mpi.slave_get_task('main slave loop')

            if task == 'init':  # data: micro_file, n_micro
                output.set_output(filename='homog_app_mpi_%d.log' % rank,
                                  quiet=True)
                micro_file, n_micro = data[:2]
                required, other = get_standard_keywords()
                required.remove('equations')
                conf = ProblemConf.from_file(micro_file, required, other,
                                             verbose=False)
                options = Struct(output_filename_trunk=None)
                homogen_app = HomogenizationApp(conf, options, 'micro:',
                                                n_micro=n_micro)
            elif task == 'calculate':  # data: rel_def_grad, ts, iteration
                macro_data, ts, iteration = data[:3]
                homogen_app.setup_macro_data(macro_data)
                homogen_app(ret_all=True, itime=ts.step, iiter=iteration)
            elif task == 'finalize':
                done = True


if __name__ == '__main__':
    main()



# -*- Mode: Python -*-
cimport cython

import numpy as np
cimport numpy as np

from sfepy.discrete.common.extmods.types cimport int32, uint32, float64

cdef extern from 'contres.h':
     cdef void _getLongestEdgeAndGPs \
          'getLongestEdgeAndGPs'(float64* longestEdge, float64* GPs,
                                 int n, int nsd, int ngp, int neq, int nsn,
                                 int nes, int nen,
                                 uint32* elementID, uint32* segmentID,
                                 int32* ISN, int32* IEN, float64* H, float64* X)
     cdef void _getAABB \
          'getAABB'(float64* AABBmin, float64* AABBmax, int nsd, int nnod,
                    float64* X, float64 longestEdge, int32* IEN, int32* ISN,
                    uint32* elementID, uint32* segmentID,
                    int n, int nsn, int nes, int nen, int neq)

     cdef void _evaluateContactConstraints \
          'evaluateContactConstraints'(float64* GPs, int32* ISN, int32* IEN,
                                       int32* N, float64* AABBmin,
                                       float64* AABBmax, int32* head,
                                       int32* next, float64* X,
                                       uint32* elementID, uint32* segmentID,
                                       int n, int nsn, int nsd, int npd,
                                       int ngp, int nen, int nes, int neq,
                                       float64 longestEdge)
     cdef void _assembleContactResidualAndStiffness \
          'assembleContactResidualAndStiffness'(
          float64* Gc, float64* vals, int32* rows, int32* cols,
          int* len, float64* GPs, int32* ISN,
          int32* IEN, float64* X, float64* U, float64* H, float64* dH,
          float64* gw, float64* activeGPsOld, int neq, int nsd, int npd,
          int ngp, int nes, int nsn, int nen, int GPs_len, float64 epss,
          int keyContactDetection, int keyAssembleKc)

def get_longest_edge_and_gps(
    np.ndarray[float64, mode='fortran', ndim=2] GPs not None,
    int neq,
    np.ndarray[uint32, mode='c', ndim=1] elementID not None,
    np.ndarray[uint32, mode='c', ndim=1] segmentID not None,
    np.ndarray[int32, mode='c', ndim=2] ISN not None,
    np.ndarray[int32, mode='c', ndim=2] IEN not None,
    np.ndarray[float64, mode='fortran', ndim=2] H not None,
    np.ndarray[float64, mode='fortran', ndim=2] X not None,
    ):

    cdef float64 tmp[1]
    cdef int n, nsd, ngp, nsn, nes, nen

    n = elementID.shape[0]
    nsd = X.shape[1]
    ngp = H.shape[0]
    nsn = ISN.shape[0]
    nes = ISN.shape[1]
    nen = IEN.shape[1]

    _getLongestEdgeAndGPs(<float64 *> &tmp, &GPs[0, 0],
                          n, nsd, ngp, neq, nsn, nes, nen,
                          &elementID[0], &segmentID[0], &ISN[0, 0], &IEN[0, 0],
                          &H[0, 0], &X[0, 0])

    longestEdge = tmp[0]

    return longestEdge, GPs

def get_AABB(
    np.ndarray[float64, mode='fortran', ndim=2] X not None,
    int longestEdge,
    np.ndarray[int32, mode='c', ndim=2] IEN not None,
    np.ndarray[int32, mode='c', ndim=2] ISN not None,
    np.ndarray[uint32, mode='c', ndim=1] elementID not None,
    np.ndarray[uint32, mode='c', ndim=1] segmentID not None,
    int neq,
    ):

    cdef np.ndarray[float64, mode='c', ndim=1] AABBmin
    cdef np.ndarray[float64, mode='c', ndim=1] AABBmax
    cdef int n, nnod, nsd, nsn, nes, nen

    n = elementID.shape[0]
    nnod = X.shape[0]
    nsd = X.shape[1]
    nsn = ISN.shape[0]
    nes = ISN.shape[1]
    nen = IEN.shape[1]

    AABBmin = np.empty((nsd,), dtype=np.float64)
    AABBmax = np.empty((nsd,), dtype=np.float64)

    _getAABB(&AABBmin[0], &AABBmax[0], nsd, nnod,
             &X[0, 0], longestEdge, &IEN[0, 0], &ISN[0, 0],
             &elementID[0], &segmentID[0],
             n, nsn, nes, nen, neq)

    return AABBmin, AABBmax

def init_global_search(
    np.ndarray[int32, mode='c', ndim=1] N not None,
    np.ndarray[float64, mode='c', ndim=1] AABBmin not None,
    np.ndarray[float64, mode='c', ndim=1] AABBmax not None,
    np.ndarray[float64, mode='fortran', ndim=2] X not None,
    ):
    """
    The linked list initialization. The head array contains, at the position
    Ic, the index of the first point that belongs to the cell Ic, the second
    point index is then next[head[Ic]], the third point index is
    next[next[head[Ic]]] etc. - the next array points from the i-th point in
    each cell to the (i+1)-th point, until -1 is reached.
    """
    cdef int i, lower, upper, Ic
    cdef np.ndarray[int32, mode='c', ndim=1] I
    cdef np.ndarray j
    cdef np.ndarray[int32, mode='c', ndim=1] head
    cdef np.ndarray[int32, mode='c', ndim=1] next

    dim = X.shape[1]
    nnod = X.shape[0]

    head = np.empty((np.prod(N),), dtype=np.int32);
    head[:] = -1

    next = np.empty((nnod,), dtype=np.int32);
    next[:] = -1

    # Loop over points to find a cell each point is in.
    for i in range(nnod):
        lower = np.sum(X[i] < AABBmin)
        higher = np.sum(X[i] > AABBmax)

        if lower + higher: continue

        I = np.floor(N * (X[i] - AABBmin) / (AABBmax-AABBmin)).astype(np.int32)

        j = I == N
        I[j] = N[j] - 1

        Ic = np.ravel_multi_index(I, N, order='F')

        next[i] = head[Ic]
        head[Ic] = i

    return head, next

def evaluate_contact_constraints(
    np.ndarray[float64, mode='fortran', ndim=2] GPs not None,
    np.ndarray[int32, mode='c', ndim=2] ISN not None,
    np.ndarray[int32, mode='c', ndim=2] IEN not None,
    np.ndarray[int32, mode='c', ndim=1] N not None,
    np.ndarray[float64, mode='c', ndim=1] AABBmin not None,
    np.ndarray[float64, mode='c', ndim=1] AABBmax not None,
    np.ndarray[int32, mode='c', ndim=1] head not None,
    np.ndarray[int32, mode='c', ndim=1] next not None,
    np.ndarray[float64, mode='fortran', ndim=2] X not None,
    np.ndarray[uint32, mode='c', ndim=1] elementID not None,
    np.ndarray[uint32, mode='c', ndim=1] segmentID not None,
    int npd,
    int neq,
    float64 longestEdge,
    ):

    cdef int n, nsd, ngp, nsn, nes, nen

    n = elementID.shape[0]
    nsd = X.shape[1]
    ngp = GPs.shape[0] / n
    nsn = ISN.shape[0]
    nes = ISN.shape[1]
    nen = IEN.shape[1]

    _evaluateContactConstraints(&GPs[0, 0], &ISN[0, 0], &IEN[0, 0], &N[0],
                                &AABBmin[0], &AABBmax[0], &head[0], &next[0],
                                &X[0, 0], &elementID[0], &segmentID[0],
                                n, nsn, nsd, npd, ngp, nen, nes, neq,
                                longestEdge)

    return GPs

def assemble_contact_residual_and_stiffness(
    np.ndarray[float64, mode='c', ndim=1] Gc not None,
    np.ndarray[float64, mode='c', ndim=1] vals not None,
    np.ndarray[int32, mode='c', ndim=1] rows not None,
    np.ndarray[int32, mode='c', ndim=1] cols not None,
    np.ndarray[float64, mode='fortran', ndim=2] GPs not None,
    np.ndarray[int32, mode='c', ndim=2] ISN not None,
    np.ndarray[int32, mode='c', ndim=2] IEN not None,
    np.ndarray[float64, mode='fortran', ndim=2] X not None,
    np.ndarray[float64, mode='fortran', ndim=2] Um not None,
    np.ndarray[float64, mode='fortran', ndim=2] H not None,
    np.ndarray[float64, mode='fortran', ndim=2] dH not None,
    np.ndarray[float64, mode='c', ndim=1] gw not None,
    np.ndarray[float64, mode='fortran', ndim=1] activeGPsOld not None,
    int neq,
    int npd,
    float64 epss,
    int keyContactDetection,
    int keyAssembleKc,
    ):
    cdef int num
    cdef int nsd, ngp, nsn, nes, nen, GPs_len

    num = len(vals)
    GPs_len = GPs.shape[0]
    nsd = X.shape[1]
    ngp = H.shape[0]
    nsn = ISN.shape[0]
    nes = ISN.shape[1]
    nen = IEN.shape[1]

    _assembleContactResidualAndStiffness(&Gc[0],
                                         &vals[0], &rows[0], &cols[0], &num,
                                         &GPs[0, 0],
                                         &ISN[0, 0], &IEN[0, 0], &X[0, 0],
                                         &Um[0, 0], &H[0, 0], &dH[0, 0],
                                         &gw[0], &activeGPsOld[0],
                                         neq, nsd, npd, ngp, nes, nsn, nen,
                                         GPs_len, epss,
                                         keyContactDetection, keyAssembleKc)

    return Gc, vals, rows, cols, num
